@@ -3,12 +3,12 @@
  * hashovfl.c
  *	  Overflow page management code for the Postgres hash access method
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/access/hash/hashovfl.c,v 1.27 2000/03/17 02:36:02 tgl Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/access/hash/hashovfl.c,v 1.29 2001/03/07 21:20:26 tgl Exp $
  *
  * NOTES
  *	  Overflow pages look like ordinary relation pages.
@@ -564,7 +564,10 @@ _hash_squeezebucket(Relation rel,
 		 * page.
 		 */
 		woffnum = OffsetNumberNext(PageGetMaxOffsetNumber(wpage));
-		PageAddItem(wpage, (Item) hitem, itemsz, woffnum, LP_USED);
+		if (PageAddItem(wpage, (Item) hitem, itemsz, woffnum, LP_USED)
+			== InvalidOffsetNumber)
+			elog(ERROR, "_hash_squeezebucket: failed to add index item to %s",
+				 RelationGetRelationName(rel));
 
 		/*
 		 * delete the tuple from the "read" page. PageIndexTupleDelete
