@@ -3,11 +3,12 @@
  * btutils.c
  *	  Utility code for Postgres btree implementation.
  *
- * Copyright (c) 1994, Regents of the University of California
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/access/hash/hashutil.c,v 1.15.2.1 1999/08/02 05:24:36 scrappy Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/access/hash/hashutil.c,v 1.26 2001/02/22 21:48:49 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -71,7 +72,7 @@ _hash_formitem(IndexTuple itup)
 	Size		tuplen;
 
 	/* disallow nulls in hash keys */
-	if (itup->t_info & INDEX_NULL_MASK)
+	if (IndexTupleHasNulls(itup))
 		elog(ERROR, "hash indices cannot include null keys");
 
 	/* make a copy of the index tuple with room for the sequence number */
@@ -90,10 +91,8 @@ _hash_call(Relation rel, HashMetaPage metap, Datum key)
 {
 	uint32		n;
 	Bucket		bucket;
-	RegProcedure proc;
 
-	proc = metap->hashm_procid;
-	n = (uint32) fmgr(proc, key);
+	n = DatumGetUInt32(OidFunctionCall1(metap->hashm_procid, key));
 	bucket = n & metap->hashm_highmask;
 	if (bucket > metap->hashm_maxbucket)
 		bucket = bucket & metap->hashm_lowmask;
