@@ -2,23 +2,23 @@
  *
  * copy.c
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/commands/copy.c,v 1.105 2000/04/16 04:27:52 tgl Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/commands/copy.c,v 1.136 2001/03/22 06:16:11 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
+#include "postgres.h"
 
 #include <unistd.h>
 #include <sys/stat.h>
 
-#include "postgres.h"
-
 #include "access/genam.h"
 #include "access/heapam.h"
+#include "access/printtup.h"
 #include "catalog/catname.h"
 #include "catalog/index.h"
 #include "catalog/pg_index.h"
@@ -27,12 +27,13 @@
 #include "commands/copy.h"
 #include "commands/trigger.h"
 #include "executor/executor.h"
-#include "lib/stringinfo.h"
 #include "libpq/libpq.h"
 #include "miscadmin.h"
+#include "tcop/pquery.h"
 #include "tcop/tcopprot.h"
 #include "utils/acl.h"
 #include "utils/builtins.h"
+#include "utils/relcache.h"
 #include "utils/syscache.h"
 
 #ifdef MULTIBYTE
@@ -46,25 +47,19 @@
 /* non-export function prototypes */
 static void CopyTo(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null_print);
 static void CopyFrom(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null_print);
-static Oid	GetOutputFunction(Oid type);
-static Oid	GetTypeElement(Oid type);
 static Oid	GetInputFunction(Oid type);
-static Oid	IsTypeByVal(Oid type);
-static void GetIndexRelations(Oid main_relation_oid,
-				  int *n_indices,
-				  Relation **index_rels);
-
+static Oid	GetTypeElement(Oid type);
 static void CopyReadNewline(FILE *fp, int *newline);
 static char *CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_print);
-
 static void CopyAttributeOut(FILE *fp, char *string, char *delim);
-static int	CountTuples(Relation relation);
+
+static const char BinarySignature[12] = "PGBCOPY\n\377\r\n\0";
 
 /*
  * Static communication variables ... pretty grotty, but COPY has
  * never been reentrant...
  */
-int			lineno = 0;			/* used by elog() -- dz */
+int			lineno = 0;			/* exported for use by elog() -- dz */
 static bool fe_eof;
 
 /*
@@ -79,7 +74,8 @@ static bool fe_eof;
 static StringInfoData attribute_buf;
 
 #ifdef MULTIBYTE
-static int	encoding;
+static int	client_encoding;
+static int	server_encoding;
 
 #endif
 
@@ -275,20 +271,15 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 
 	FILE	   *fp;
 	Relation	rel;
-	extern char *UserName;		/* defined in global.c */
 	const AclMode required_access = from ? ACL_WR : ACL_RD;
 	int			result;
 
 	/*
 	 * Open and lock the relation, using the appropriate lock type.
-	 *
-	 * Note: AccessExclusive is probably overkill for copying to a relation,
-	 * but that's what the code grabs on the rel's indices.  If this lock
-	 * is relaxed then I think the index locks need relaxed also.
 	 */
-	rel = heap_openr(relname, (from ? AccessExclusiveLock : AccessShareLock));
+	rel = heap_openr(relname, (from ? RowExclusiveLock : AccessShareLock));
 
-	result = pg_aclcheck(relname, UserName, required_access);
+	result = pg_aclcheck(relname, GetUserId(), required_access);
 	if (result != ACLCHECK_OK)
 		elog(ERROR, "%s: %s", relname, aclcheck_error_strings[result]);
 	if (!pipe && !superuser())
@@ -296,18 +287,26 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 			 "directly to or from a file.  Anyone can COPY to stdout or "
 			 "from stdin.  Psql's \\copy command also works for anyone.");
 
+	/*
+	 * This restriction is unfortunate, but necessary until the frontend
+	 * COPY protocol is redesigned to be binary-safe...
+	 */
+	if (pipe && binary)
+		elog(ERROR, "COPY BINARY is not supported to stdout or from stdin");
+
 	/*
 	 * Set up variables to avoid per-attribute overhead.
 	 */
 	initStringInfo(&attribute_buf);
 #ifdef MULTIBYTE
-	encoding = pg_get_client_encoding();
+	client_encoding = pg_get_client_encoding();
+	server_encoding = GetDatabaseEncoding();
 #endif
 
 	if (from)
 	{							/* copy from file to database */
 		if (rel->rd_rel->relkind == RELKIND_SEQUENCE)
-			elog(ERROR, "You can't change sequence relation %s", relname);
+			elog(ERROR, "You cannot change sequence relation %s", relname);
 		if (pipe)
 		{
 			if (IsUnderPostmaster)
@@ -320,16 +319,12 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 		}
 		else
 		{
-#ifndef __CYGWIN32__
-			fp = AllocateFile(filename, "r");
-#else
-			fp = AllocateFile(filename, "rb");
-#endif
+			fp = AllocateFile(filename, PG_BINARY_R);
 			if (fp == NULL)
 				elog(ERROR, "COPY command, running in backend with "
 					 "effective uid %d, could not open file '%s' for "
 					 "reading.  Errno = %s (%d).",
-					 geteuid(), filename, strerror(errno), errno);
+					 (int) geteuid(), filename, strerror(errno), errno);
 		}
 		CopyFrom(rel, binary, oids, fp, delim, null_print);
 	}
@@ -350,24 +345,23 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 		{
 			mode_t		oumask; /* Pre-existing umask value */
 
-			if (*filename != '/')
+			/*
+			 * Prevent write to relative path ... too easy to shoot
+			 * oneself in the foot by overwriting a database file ...
+			 */
+			if (filename[0] != '/')
 				elog(ERROR, "Relative path not allowed for server side"
 					 " COPY command.");
 
 			oumask = umask((mode_t) 022);
-
-#ifndef __CYGWIN32__
-			fp = AllocateFile(filename, "w");
-#else
-			fp = AllocateFile(filename, "wb");
-#endif
+			fp = AllocateFile(filename, PG_BINARY_W);
 			umask(oumask);
 
 			if (fp == NULL)
 				elog(ERROR, "COPY command, running in backend with "
 					 "effective uid %d, could not open file '%s' for "
 					 "writing.  Errno = %s (%d).",
-					 geteuid(), filename, strerror(errno), errno);
+					 (int) geteuid(), filename, strerror(errno), errno);
 		}
 		CopyTo(rel, binary, oids, fp, delim, null_print);
 	}
@@ -393,438 +387,377 @@ DoCopy(char *relname, bool binary, bool oids, bool from, bool pipe,
 }
 
 
-
+/*
+ * Copy from relation TO file.
+ */
 static void
-CopyTo(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null_print)
+CopyTo(Relation rel, bool binary, bool oids, FILE *fp,
+	   char *delim, char *null_print)
 {
 	HeapTuple	tuple;
+	TupleDesc	tupDesc;
 	HeapScanDesc scandesc;
-
-	int32		attr_count,
+	int			attr_count,
 				i;
-
-#ifdef	_DROP_COLUMN_HACK__
-	bool	   *valid;
-
-#endif	 /* _DROP_COLUMN_HACK__ */
 	Form_pg_attribute *attr;
 	FmgrInfo   *out_functions;
-	Oid			out_func_oid;
 	Oid		   *elements;
-	int32	   *typmod;
-	Datum		value;
-	bool		isnull;			/* The attribute we are copying is null */
-	char	   *nulls;
+	bool	   *isvarlena;
+	int16		fld_size;
+	char	   *string;
+
+	tupDesc = rel->rd_att;
+	attr_count = rel->rd_att->natts;
+	attr = rel->rd_att->attrs;
 
 	/*
-	 * <nulls> is a (dynamically allocated) array with one character per
-	 * attribute in the instance being copied.	nulls[I-1] is 'n' if
-	 * Attribute Number I is null, and ' ' otherwise.
-	 *
-	 * <nulls> is meaningful only if we are doing a binary copy.
+	 * For binary copy we really only need isvarlena, but compute it
+	 * all...
 	 */
-	char	   *string;
-	int32		ntuples;
-	TupleDesc	tupDesc;
+	out_functions = (FmgrInfo *) palloc(attr_count * sizeof(FmgrInfo));
+	elements = (Oid *) palloc(attr_count * sizeof(Oid));
+	isvarlena = (bool *) palloc(attr_count * sizeof(bool));
+	for (i = 0; i < attr_count; i++)
+	{
+		Oid			out_func_oid;
+
+		if (!getTypeOutputInfo(attr[i]->atttypid,
+							 &out_func_oid, &elements[i], &isvarlena[i]))
+			elog(ERROR, "COPY: couldn't lookup info for type %u",
+				 attr[i]->atttypid);
+		fmgr_info(out_func_oid, &out_functions[i]);
+	}
+
+	if (binary)
+	{
+		/* Generate header for a binary copy */
+		int32		tmp;
+
+		/* Signature */
+		CopySendData((char *) BinarySignature, 12, fp);
+		/* Integer layout field */
+		tmp = 0x01020304;
+		CopySendData(&tmp, sizeof(int32), fp);
+		/* Flags field */
+		tmp = 0;
+		if (oids)
+			tmp |= (1 << 16);
+		CopySendData(&tmp, sizeof(int32), fp);
+		/* No header extension */
+		tmp = 0;
+		CopySendData(&tmp, sizeof(int32), fp);
+	}
 
 	scandesc = heap_beginscan(rel, 0, QuerySnapshot, 0, NULL);
 
-	attr_count = rel->rd_att->natts;
-	attr = rel->rd_att->attrs;
-	tupDesc = rel->rd_att;
-
-	if (!binary)
-	{
-		out_functions = (FmgrInfo *) palloc(attr_count * sizeof(FmgrInfo));
-		elements = (Oid *) palloc(attr_count * sizeof(Oid));
-		typmod = (int32 *) palloc(attr_count * sizeof(int32));
-#ifdef	_DROP_COLUMN_HACK__
-		valid = (bool *) palloc(attr_count * sizeof(bool));
-#endif	 /* _DROP_COLUMN_HACK__ */
-		for (i = 0; i < attr_count; i++)
-		{
-#ifdef	_DROP_COLUMN_HACK__
-			if (COLUMN_IS_DROPPED(attr[i]))
-			{
-				valid[i] = false;
-				continue;
-			}
-			else
-				valid[i] = true;
-#endif	 /* _DROP_COLUMN_HACK__ */
-			out_func_oid = (Oid) GetOutputFunction(attr[i]->atttypid);
-			fmgr_info(out_func_oid, &out_functions[i]);
-			elements[i] = GetTypeElement(attr[i]->atttypid);
-			typmod[i] = attr[i]->atttypmod;
-		}
-		nulls = NULL;			/* meaningless, but compiler doesn't know
-								 * that */
-	}
-	else
-	{
-		elements = NULL;
-		typmod = NULL;
-		out_functions = NULL;
-		nulls = (char *) palloc(attr_count);
-		for (i = 0; i < attr_count; i++)
-			nulls[i] = ' ';
-
-		/* XXX expensive */
-
-		ntuples = CountTuples(rel);
-		CopySendData(&ntuples, sizeof(int32), fp);
-	}
-
 	while (HeapTupleIsValid(tuple = heap_getnext(scandesc, 0)))
 	{
-		if (QueryCancel)
-			CancelQuery();
+		bool		need_delim = false;
 
-		if (oids && !binary)
+		CHECK_FOR_INTERRUPTS();
+
+		if (binary)
 		{
-			CopySendString(oidout(tuple->t_data->t_oid), fp);
-			CopySendChar(delim[0], fp);
+			/* Binary per-tuple header */
+			int16		fld_count = attr_count;
+
+			CopySendData(&fld_count, sizeof(int16), fp);
+			/* Send OID if wanted --- note fld_count doesn't include it */
+			if (oids)
+			{
+				fld_size = sizeof(Oid);
+				CopySendData(&fld_size, sizeof(int16), fp);
+				CopySendData(&tuple->t_data->t_oid, sizeof(Oid), fp);
+			}
+		}
+		else
+		{
+			/* Text format has no per-tuple header, but send OID if wanted */
+			if (oids)
+			{
+				string = DatumGetCString(DirectFunctionCall1(oidout,
+								ObjectIdGetDatum(tuple->t_data->t_oid)));
+				CopySendString(string, fp);
+				pfree(string);
+				need_delim = true;
+			}
 		}
 
 		for (i = 0; i < attr_count; i++)
 		{
-			value = heap_getattr(tuple, i + 1, tupDesc, &isnull);
+			Datum		origvalue,
+						value;
+			bool		isnull;
+
+			origvalue = heap_getattr(tuple, i + 1, tupDesc, &isnull);
+
 			if (!binary)
 			{
-#ifdef	_DROP_COLUMN_HACK__
-				if (!valid[i])
-				{
-					if (i == attr_count - 1)
-						CopySendChar('\n', fp);
-					continue;
-				}
-#endif	 /* _DROP_COLUMN_HACK__ */
-				if (!isnull)
-				{
-					string = (char *) (*fmgr_faddr(&out_functions[i]))
-						(value, elements[i], typmod[i]);
-					CopyAttributeOut(fp, string, delim);
-					pfree(string);
-				}
-				else
-					CopySendString(null_print, fp);		/* null indicator */
-
-				if (i == attr_count - 1)
-					CopySendChar('\n', fp);
-				else
-				{
-
-					/*
-					 * when copying out, only use the first char of the
-					 * delim string
-					 */
+				if (need_delim)
 					CopySendChar(delim[0], fp);
+				need_delim = true;
+			}
+
+			if (isnull)
+			{
+				if (!binary)
+				{
+					CopySendString(null_print, fp);		/* null indicator */
+				}
+				else
+				{
+					fld_size = 0;		/* null marker */
+					CopySendData(&fld_size, sizeof(int16), fp);
 				}
 			}
 			else
 			{
 
 				/*
-				 * only interesting thing heap_getattr tells us in this
-				 * case is if we have a null attribute or not.
+				 * If we have a toasted datum, forcibly detoast it to
+				 * avoid memory leakage inside the type's output routine
+				 * (or for binary case, becase we must output untoasted
+				 * value).
 				 */
-				if (isnull)
-					nulls[i] = 'n';
-			}
-		}
+				if (isvarlena[i])
+					value = PointerGetDatum(PG_DETOAST_DATUM(origvalue));
+				else
+					value = origvalue;
 
-		if (binary)
-		{
-			int32		null_ct = 0,
-						length;
-
-			for (i = 0; i < attr_count; i++)
-			{
-				if (nulls[i] == 'n')
-					null_ct++;
-			}
-
-			length = tuple->t_len - tuple->t_data->t_hoff;
-			CopySendData(&length, sizeof(int32), fp);
-			if (oids)
-				CopySendData((char *) &tuple->t_data->t_oid, sizeof(int32), fp);
-
-			CopySendData(&null_ct, sizeof(int32), fp);
-			if (null_ct > 0)
-			{
-				for (i = 0; i < attr_count; i++)
+				if (!binary)
 				{
-					if (nulls[i] == 'n')
+					string = DatumGetCString(FunctionCall3(&out_functions[i],
+														   value,
+										   ObjectIdGetDatum(elements[i]),
+									 Int32GetDatum(attr[i]->atttypmod)));
+					CopyAttributeOut(fp, string, delim);
+					pfree(string);
+				}
+				else
+				{
+					fld_size = attr[i]->attlen;
+					CopySendData(&fld_size, sizeof(int16), fp);
+					if (isvarlena[i])
 					{
-						CopySendData(&i, sizeof(int32), fp);
-						nulls[i] = ' ';
+						/* varlena */
+						Assert(fld_size == -1);
+						CopySendData(DatumGetPointer(value),
+									 VARSIZE(value),
+									 fp);
+					}
+					else if (!attr[i]->attbyval)
+					{
+						/* fixed-length pass-by-reference */
+						Assert(fld_size > 0);
+						CopySendData(DatumGetPointer(value),
+									 fld_size,
+									 fp);
+					}
+					else
+					{
+						/* pass-by-value */
+						Datum		datumBuf;
+
+						/*
+						 * We need this horsing around because we don't
+						 * know how shorter data values are aligned within
+						 * a Datum.
+						 */
+						store_att_byval(&datumBuf, value, fld_size);
+						CopySendData(&datumBuf,
+									 fld_size,
+									 fp);
 					}
 				}
+
+				/* Clean up detoasted copy, if any */
+				if (value != origvalue)
+					pfree(DatumGetPointer(value));
 			}
-			CopySendData((char *) tuple->t_data + tuple->t_data->t_hoff,
-						 length, fp);
 		}
+
+		if (!binary)
+			CopySendChar('\n', fp);
 	}
 
 	heap_endscan(scandesc);
+
 	if (binary)
-		pfree(nulls);
-	else
 	{
-		pfree(out_functions);
-		pfree(elements);
-		pfree(typmod);
+		/* Generate trailer for a binary copy */
+		int16		fld_count = -1;
+
+		CopySendData(&fld_count, sizeof(int16), fp);
 	}
+
+	pfree(out_functions);
+	pfree(elements);
+	pfree(isvarlena);
 }
 
+
+/*
+ * Copy FROM file to relation.
+ */
 static void
-CopyFrom(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null_print)
+CopyFrom(Relation rel, bool binary, bool oids, FILE *fp,
+		 char *delim, char *null_print)
 {
 	HeapTuple	tuple;
-	AttrNumber	attr_count;
+	TupleDesc	tupDesc;
 	Form_pg_attribute *attr;
+	AttrNumber	attr_count;
 	FmgrInfo   *in_functions;
+	Oid		   *elements;
 	int			i;
 	Oid			in_func_oid;
 	Datum	   *values;
-	char	   *nulls,
-			   *index_nulls;
-	bool	   *byval;
+	char	   *nulls;
 	bool		isnull;
-	bool		has_index;
 	int			done = 0;
-	char	   *string = NULL,
-			   *ptr;
-	Relation   *index_rels;
-	int32		len,
-				null_ct,
-				null_id;
-	int32		ntuples,
-				tuples_read = 0;
-	bool		reading_to_eof = true;
-	Oid		   *elements;
-	int32	   *typmod;
-	FuncIndexInfo *finfo,
-			  **finfoP = NULL;
-	TupleDesc  *itupdescArr;
-	HeapTuple	pgIndexTup;
-	Form_pg_index *pgIndexP = NULL;
-	int		   *indexNatts = NULL;
-	char	   *predString;
-	Node	  **indexPred = NULL;
-	TupleDesc	rtupdesc;
-	ExprContext *econtext = NULL;
-	EState	   *estate = makeNode(EState);		/* for ExecConstraints() */
-
-#ifndef OMIT_PARTIAL_INDEX
+	char	   *string;
+	ResultRelInfo *resultRelInfo;
+	EState	   *estate = CreateExecutorState(); /* for ExecConstraints() */
 	TupleTable	tupleTable;
-	TupleTableSlot *slot = NULL;
-
-#endif
-	int			natts;
-	AttrNumber *attnumP;
-	Datum	   *idatum;
-	int			n_indices;
-	InsertIndexResult indexRes;
-	TupleDesc	tupDesc;
+	TupleTableSlot *slot;
 	Oid			loaded_oid = InvalidOid;
 	bool		skip_tuple = false;
+	bool		file_has_oids;
 
 	tupDesc = RelationGetDescr(rel);
 	attr = tupDesc->attrs;
 	attr_count = tupDesc->natts;
 
-	has_index = false;
-
 	/*
-	 * This may be a scalar or a functional index.	We initialize all
-	 * kinds of arrays here to avoid doing extra work at every tuple copy.
+	 * We need a ResultRelInfo so we can use the regular executor's
+	 * index-entry-making machinery.  (There used to be a huge amount of
+	 * code here that basically duplicated execUtils.c ...)
 	 */
+	resultRelInfo = makeNode(ResultRelInfo);
+	resultRelInfo->ri_RangeTableIndex = 1;		/* dummy */
+	resultRelInfo->ri_RelationDesc = rel;
 
-	if (rel->rd_rel->relhasindex)
-	{
-		GetIndexRelations(RelationGetRelid(rel), &n_indices, &index_rels);
-		if (n_indices > 0)
-		{
-			has_index = true;
-			itupdescArr = (TupleDesc *) palloc(n_indices * sizeof(TupleDesc));
-			pgIndexP = (Form_pg_index *) palloc(n_indices * sizeof(Form_pg_index));
-			indexNatts = (int *) palloc(n_indices * sizeof(int));
-			finfo = (FuncIndexInfo *) palloc(n_indices * sizeof(FuncIndexInfo));
-			finfoP = (FuncIndexInfo **) palloc(n_indices * sizeof(FuncIndexInfo *));
-			indexPred = (Node **) palloc(n_indices * sizeof(Node *));
-			econtext = NULL;
-			for (i = 0; i < n_indices; i++)
-			{
-				itupdescArr[i] = RelationGetDescr(index_rels[i]);
-				pgIndexTup = SearchSysCacheTuple(INDEXRELID,
-					   ObjectIdGetDatum(RelationGetRelid(index_rels[i])),
-												 0, 0, 0);
-				Assert(pgIndexTup);
-				pgIndexP[i] = (Form_pg_index) GETSTRUCT(pgIndexTup);
-				for (attnumP = &(pgIndexP[i]->indkey[0]), natts = 0;
-				 natts < INDEX_MAX_KEYS && *attnumP != InvalidAttrNumber;
-					 attnumP++, natts++);
-				if (pgIndexP[i]->indproc != InvalidOid)
-				{
-					FIgetnArgs(&finfo[i]) = natts;
-					natts = 1;
-					FIgetProcOid(&finfo[i]) = pgIndexP[i]->indproc;
-					*(FIgetname(&finfo[i])) = '\0';
-					finfoP[i] = &finfo[i];
-				}
-				else
-					finfoP[i] = (FuncIndexInfo *) NULL;
-				indexNatts[i] = natts;
-				if (VARSIZE(&pgIndexP[i]->indpred) != 0)
-				{
-					predString = fmgr(F_TEXTOUT, &pgIndexP[i]->indpred);
-					indexPred[i] = stringToNode(predString);
-					pfree(predString);
-					/* make dummy ExprContext for use by ExecQual */
-					if (econtext == NULL)
-					{
-#ifndef OMIT_PARTIAL_INDEX
-						tupleTable = ExecCreateTupleTable(1);
-						slot = ExecAllocTableSlot(tupleTable);
-						econtext = makeNode(ExprContext);
-						econtext->ecxt_scantuple = slot;
-						rtupdesc = RelationGetDescr(rel);
-						slot->ttc_tupleDescriptor = rtupdesc;
+	ExecOpenIndices(resultRelInfo);
 
-						/*
-						 * There's no buffer associated with heap tuples
-						 * here, so I set the slot's buffer to NULL.
-						 * Currently, it appears that the only way a
-						 * buffer could be needed would be if the partial
-						 * index predicate referred to the "lock" system
-						 * attribute.  If it did, then heap_getattr would
-						 * call HeapTupleGetRuleLock, which uses the
-						 * buffer's descriptor to get the relation id.
-						 * Rather than try to fix this, I'll just disallow
-						 * partial indexes on "lock", which wouldn't be
-						 * useful anyway. --Nels, Nov '92
-						 */
-						/* SetSlotBuffer(slot, (Buffer) NULL); */
-						/* SetSlotShouldFree(slot, false); */
-						slot->ttc_buffer = (Buffer) NULL;
-						slot->ttc_shouldFree = false;
-#endif	 /* OMIT_PARTIAL_INDEX */
-					}
-				}
-				else
-					indexPred[i] = NULL;
-			}
-		}
-	}
+	estate->es_result_relations = resultRelInfo;
+	estate->es_num_result_relations = 1;
+	estate->es_result_relation_info = resultRelInfo;
+
+	/* Set up a dummy tuple table too */
+	tupleTable = ExecCreateTupleTable(1);
+	slot = ExecAllocTableSlot(tupleTable);
+	ExecSetSlotDescriptor(slot, tupDesc, false);
 
 	if (!binary)
 	{
 		in_functions = (FmgrInfo *) palloc(attr_count * sizeof(FmgrInfo));
 		elements = (Oid *) palloc(attr_count * sizeof(Oid));
-		typmod = (int32 *) palloc(attr_count * sizeof(int32));
 		for (i = 0; i < attr_count; i++)
 		{
-#ifdef	_DROP_COLUMN_HACK__
-			if (COLUMN_IS_DROPPED(attr[i]))
-				continue;
-#endif	 /* _DROP_COLUMN_HACK__ */
 			in_func_oid = (Oid) GetInputFunction(attr[i]->atttypid);
 			fmgr_info(in_func_oid, &in_functions[i]);
 			elements[i] = GetTypeElement(attr[i]->atttypid);
-			typmod[i] = attr[i]->atttypmod;
 		}
+		file_has_oids = oids;	/* must rely on user to tell us this... */
 	}
 	else
 	{
+		/* Read and verify binary header */
+		char		readSig[12];
+		int32		tmp;
+
+		/* Signature */
+		CopyGetData(readSig, 12, fp);
+		if (CopyGetEof(fp) ||
+			memcmp(readSig, BinarySignature, 12) != 0)
+			elog(ERROR, "COPY BINARY: file signature not recognized");
+		/* Integer layout field */
+		CopyGetData(&tmp, sizeof(int32), fp);
+		if (CopyGetEof(fp) ||
+			tmp != 0x01020304)
+			elog(ERROR, "COPY BINARY: incompatible integer layout");
+		/* Flags field */
+		CopyGetData(&tmp, sizeof(int32), fp);
+		if (CopyGetEof(fp))
+			elog(ERROR, "COPY BINARY: bogus file header (missing flags)");
+		file_has_oids = (tmp & (1 << 16)) != 0;
+		tmp &= ~(1 << 16);
+		if ((tmp >> 16) != 0)
+			elog(ERROR, "COPY BINARY: unrecognized critical flags in header");
+		/* Header extension length */
+		CopyGetData(&tmp, sizeof(int32), fp);
+		if (CopyGetEof(fp) ||
+			tmp < 0)
+			elog(ERROR, "COPY BINARY: bogus file header (missing length)");
+		/* Skip extension header, if present */
+		while (tmp-- > 0)
+		{
+			CopyGetData(readSig, 1, fp);
+			if (CopyGetEof(fp))
+				elog(ERROR, "COPY BINARY: bogus file header (wrong length)");
+		}
+
 		in_functions = NULL;
 		elements = NULL;
-		typmod = NULL;
-		CopyGetData(&ntuples, sizeof(int32), fp);
-		if (ntuples != 0)
-			reading_to_eof = false;
 	}
 
-	values = (Datum *) palloc(sizeof(Datum) * attr_count);
-	nulls = (char *) palloc(attr_count);
-	index_nulls = (char *) palloc(attr_count);
-	idatum = (Datum *) palloc(sizeof(Datum) * attr_count);
-	byval = (bool *) palloc(attr_count * sizeof(bool));
-
-	for (i = 0; i < attr_count; i++)
-	{
-		nulls[i] = ' ';
-		index_nulls[i] = ' ';
-#ifdef	_DROP_COLUMN_HACK__
-		if (COLUMN_IS_DROPPED(attr[i]))
-		{
-			byval[i] = 'n';
-			continue;
-		}
-#endif	 /* _DROP_COLUMN_HACK__ */
-		byval[i] = (bool) IsTypeByVal(attr[i]->atttypid);
-	}
+	values = (Datum *) palloc(attr_count * sizeof(Datum));
+	nulls = (char *) palloc(attr_count * sizeof(char));
 
 	lineno = 0;
 	fe_eof = false;
 
 	while (!done)
 	{
-		if (QueryCancel)
-		{
-			lineno = 0;
-			CancelQuery();
-		}
+		CHECK_FOR_INTERRUPTS();
 
 		lineno++;
 
+		/* Reset the per-output-tuple exprcontext */
+		ResetPerTupleExprContext(estate);
+
+		/* Initialize all values for row to NULL */
+		MemSet(values, 0, attr_count * sizeof(Datum));
+		MemSet(nulls, 'n', attr_count * sizeof(char));
+
 		if (!binary)
 		{
 			int			newline = 0;
 
-			if (oids)
+			if (file_has_oids)
 			{
-				string = CopyReadAttribute(fp, &isnull, delim, &newline, null_print);
-				if (string == NULL)
-					done = 1;
+				string = CopyReadAttribute(fp, &isnull, delim,
+										   &newline, null_print);
+				if (isnull)
+					elog(ERROR, "COPY TEXT: NULL Oid");
+				else if (string == NULL)
+					done = 1;	/* end of file */
 				else
 				{
-					loaded_oid = oidin(string);
+					loaded_oid = DatumGetObjectId(DirectFunctionCall1(oidin,
+											   CStringGetDatum(string)));
 					if (loaded_oid == InvalidOid)
 						elog(ERROR, "COPY TEXT: Invalid Oid");
 				}
 			}
+
 			for (i = 0; i < attr_count && !done; i++)
 			{
-#ifdef	_DROP_COLUMN_HACK__
-				if (COLUMN_IS_DROPPED(attr[i]))
-				{
-					values[i] = PointerGetDatum(NULL);
-					nulls[i] = 'n';
-					continue;
-				}
-#endif	 /* _DROP_COLUMN_HACK__ */
-				string = CopyReadAttribute(fp, &isnull, delim, &newline, null_print);
+				string = CopyReadAttribute(fp, &isnull, delim,
+										   &newline, null_print);
 				if (isnull)
 				{
-					values[i] = PointerGetDatum(NULL);
-					nulls[i] = 'n';
+					/* already set values[i] and nulls[i] */
 				}
 				else if (string == NULL)
-					done = 1;
+					done = 1;	/* end of file */
 				else
 				{
-					values[i] = (Datum) (*fmgr_faddr(&in_functions[i])) (string,
-															 elements[i],
-															  typmod[i]);
-
-					/*
-					 * Sanity check - by reference attributes cannot
-					 * return NULL
-					 */
-					if (!PointerIsValid(values[i]) &&
-						!(rel->rd_att->attrs[i]->attbyval))
-						elog(ERROR, "COPY: Bad file format");
+					values[i] = FunctionCall3(&in_functions[i],
+											  CStringGetDatum(string),
+										   ObjectIdGetDatum(elements[i]),
+									  Int32GetDatum(attr[i]->atttypmod));
+					nulls[i] = ' ';
 				}
 			}
 			if (!done)
@@ -832,82 +765,115 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null
 		}
 		else
 		{						/* binary */
-			CopyGetData(&len, sizeof(int32), fp);
-			if (CopyGetEof(fp))
+			int16		fld_count,
+						fld_size;
+
+			CopyGetData(&fld_count, sizeof(int16), fp);
+			if (CopyGetEof(fp) ||
+				fld_count == -1)
 				done = 1;
 			else
 			{
-				if (oids)
+				if (fld_count <= 0 || fld_count > attr_count)
+					elog(ERROR, "COPY BINARY: tuple field count is %d, expected %d",
+						 (int) fld_count, attr_count);
+
+				if (file_has_oids)
 				{
-					CopyGetData(&loaded_oid, sizeof(int32), fp);
+					CopyGetData(&fld_size, sizeof(int16), fp);
+					if (CopyGetEof(fp))
+						elog(ERROR, "COPY BINARY: unexpected EOF");
+					if (fld_size != (int16) sizeof(Oid))
+						elog(ERROR, "COPY BINARY: sizeof(Oid) is %d, expected %d",
+							 (int) fld_size, (int) sizeof(Oid));
+					CopyGetData(&loaded_oid, sizeof(Oid), fp);
+					if (CopyGetEof(fp))
+						elog(ERROR, "COPY BINARY: unexpected EOF");
 					if (loaded_oid == InvalidOid)
 						elog(ERROR, "COPY BINARY: Invalid Oid");
 				}
-				CopyGetData(&null_ct, sizeof(int32), fp);
-				if (null_ct > 0)
+
+				for (i = 0; i < (int) fld_count; i++)
 				{
-					for (i = 0; i < null_ct; i++)
+					CopyGetData(&fld_size, sizeof(int16), fp);
+					if (CopyGetEof(fp))
+						elog(ERROR, "COPY BINARY: unexpected EOF");
+					if (fld_size == 0)
+						continue;		/* it's NULL; nulls[i] already set */
+					if (fld_size != attr[i]->attlen)
+						elog(ERROR, "COPY BINARY: sizeof(field %d) is %d, expected %d",
+						   i + 1, (int) fld_size, (int) attr[i]->attlen);
+					if (fld_size == -1)
 					{
-						CopyGetData(&null_id, sizeof(int32), fp);
-						nulls[null_id] = 'n';
+						/* varlena field */
+						int32		varlena_size;
+						Pointer		varlena_ptr;
+
+						CopyGetData(&varlena_size, sizeof(int32), fp);
+						if (CopyGetEof(fp))
+							elog(ERROR, "COPY BINARY: unexpected EOF");
+						if (varlena_size < (int32) sizeof(int32))
+							elog(ERROR, "COPY BINARY: bogus varlena length");
+						varlena_ptr = (Pointer) palloc(varlena_size);
+						VARATT_SIZEP(varlena_ptr) = varlena_size;
+						CopyGetData(VARDATA(varlena_ptr),
+									varlena_size - sizeof(int32),
+									fp);
+						if (CopyGetEof(fp))
+							elog(ERROR, "COPY BINARY: unexpected EOF");
+						values[i] = PointerGetDatum(varlena_ptr);
 					}
-				}
-
-				string = (char *) palloc(len);
-				CopyGetData(string, len, fp);
-
-				ptr = string;
-
-				for (i = 0; i < attr_count; i++)
-				{
-					if (byval[i] && nulls[i] != 'n')
+					else if (!attr[i]->attbyval)
 					{
+						/* fixed-length pass-by-reference */
+						Pointer		refval_ptr;
 
-						switch (attr[i]->attlen)
-						{
-							case sizeof(char):
-								values[i] = (Datum) *(unsigned char *) ptr;
-								ptr += sizeof(char);
-								break;
-							case sizeof(short):
-								ptr = (char *) SHORTALIGN(ptr);
-								values[i] = (Datum) *(unsigned short *) ptr;
-								ptr += sizeof(short);
-								break;
-							case sizeof(int32):
-								ptr = (char *) INTALIGN(ptr);
-								values[i] = (Datum) *(uint32 *) ptr;
-								ptr += sizeof(int32);
-								break;
-							default:
-								elog(ERROR, "COPY BINARY: impossible size");
-								break;
-						}
+						Assert(fld_size > 0);
+						refval_ptr = (Pointer) palloc(fld_size);
+						CopyGetData(refval_ptr, fld_size, fp);
+						if (CopyGetEof(fp))
+							elog(ERROR, "COPY BINARY: unexpected EOF");
+						values[i] = PointerGetDatum(refval_ptr);
 					}
-					else if (nulls[i] != 'n')
+					else
 					{
-						ptr = (char *) att_align(ptr, attr[i]->attlen, attr[i]->attalign);
-						values[i] = (Datum) ptr;
-						ptr = att_addlength(ptr, attr[i]->attlen, ptr);
+						/* pass-by-value */
+						Datum		datumBuf;
+
+						/*
+						 * We need this horsing around because we don't
+						 * know how shorter data values are aligned within
+						 * a Datum.
+						 */
+						Assert(fld_size > 0 && fld_size <= sizeof(Datum));
+						CopyGetData(&datumBuf, fld_size, fp);
+						if (CopyGetEof(fp))
+							elog(ERROR, "COPY BINARY: unexpected EOF");
+						values[i] = fetch_att(&datumBuf, true, fld_size);
 					}
+
+					nulls[i] = ' ';
 				}
 			}
 		}
+
 		if (done)
-			continue;
+			break;
 
 		tuple = heap_formtuple(tupDesc, values, nulls);
-		if (oids)
+
+		if (oids && file_has_oids)
 			tuple->t_data->t_oid = loaded_oid;
 
 		skip_tuple = false;
+
 		/* BEFORE ROW INSERT Triggers */
 		if (rel->trigdesc &&
 			rel->trigdesc->n_before_row[TRIGGER_EVENT_INSERT] > 0)
 		{
 			HeapTuple	newtuple;
 
-			newtuple = ExecBRInsertTriggers(rel, tuple);
+			newtuple = ExecBRInsertTriggers(estate, rel, tuple);
 
 			if (newtuple == NULL)		/* "do nothing" */
 				skip_tuple = true;
@@ -920,250 +886,88 @@ CopyFrom(Relation rel, bool binary, bool oids, FILE *fp, char *delim, char *null
 
 		if (!skip_tuple)
 		{
-			/* ----------------
-			 * Check the constraints of a tuple
-			 * ----------------
+			ExecStoreTuple(tuple, slot, InvalidBuffer, false);
+
+			/*
+			 * Check the constraints of the tuple
 			 */
-
 			if (rel->rd_att->constr)
-				ExecConstraints("CopyFrom", rel, tuple, estate);
+				ExecConstraints("CopyFrom", resultRelInfo, slot, estate);
 
+			/*
+			 * OK, store the tuple and create index entries for it
+			 */
 			heap_insert(rel, tuple);
 
-			if (has_index)
-			{
-				for (i = 0; i < n_indices; i++)
-				{
-					if (indexPred[i] != NULL)
-					{
-#ifndef OMIT_PARTIAL_INDEX
+			if (resultRelInfo->ri_NumIndices > 0)
+				ExecInsertIndexTuples(slot, &(tuple->t_self), estate, false);
 
-						/*
-						 * if tuple doesn't satisfy predicate, don't
-						 * update index
-						 */
-						slot->val = tuple;
-						/* SetSlotContents(slot, tuple); */
-						if (!ExecQual((List *) indexPred[i], econtext, false))
-							continue;
-#endif	 /* OMIT_PARTIAL_INDEX */
-					}
-					FormIndexDatum(indexNatts[i],
-								(AttrNumber *) &(pgIndexP[i]->indkey[0]),
-								   tuple,
-								   tupDesc,
-								   idatum,
-								   index_nulls,
-								   finfoP[i]);
-					indexRes = index_insert(index_rels[i], idatum, index_nulls,
-											&(tuple->t_self), rel);
-					if (indexRes)
-						pfree(indexRes);
-				}
-			}
 			/* AFTER ROW INSERT Triggers */
-			if (rel->trigdesc &&
-				rel->trigdesc->n_after_row[TRIGGER_EVENT_INSERT] > 0)
-				ExecARInsertTriggers(rel, tuple);
+			if (rel->trigdesc)
+				ExecARInsertTriggers(estate, rel, tuple);
 		}
 
-		if (binary)
-			pfree(string);
-
 		for (i = 0; i < attr_count; i++)
 		{
-			if (!byval[i] && nulls[i] != 'n')
-			{
-				if (!binary)
-					pfree((void *) values[i]);
-			}
-			else if (nulls[i] == 'n')
-				nulls[i] = ' ';
+			if (!attr[i]->attbyval && nulls[i] != 'n')
+				pfree(DatumGetPointer(values[i]));
 		}
 
 		heap_freetuple(tuple);
-		tuples_read++;
-
-		if (!reading_to_eof && ntuples == tuples_read)
-			done = true;
 	}
+
+	/*
+	 * Done, clean up
+	 */
 	lineno = 0;
+
 	pfree(values);
 	pfree(nulls);
-	pfree(index_nulls);
-	pfree(idatum);
-	pfree(byval);
 
 	if (!binary)
 	{
 		pfree(in_functions);
 		pfree(elements);
-		pfree(typmod);
 	}
 
-	if (has_index)
-	{
-		for (i = 0; i < n_indices; i++)
-		{
-			if (index_rels[i] == NULL)
-				continue;
-			/* see comments in ExecOpenIndices() in execUtils.c */
-			if ((index_rels[i])->rd_rel->relam != BTREE_AM_OID &&
-				(index_rels[i])->rd_rel->relam != HASH_AM_OID)
-				UnlockRelation(index_rels[i], AccessExclusiveLock);
-			index_close(index_rels[i]);
-		}
-	}
+	ExecDropTupleTable(tupleTable, true);
+
+	ExecCloseIndices(resultRelInfo);
 }
 
 
-
 static Oid
-GetOutputFunction(Oid type)
+GetInputFunction(Oid type)
 {
 	HeapTuple	typeTuple;
+	Oid			result;
 
-	typeTuple = SearchSysCacheTuple(TYPEOID,
-									ObjectIdGetDatum(type),
-									0, 0, 0);
-
-	if (HeapTupleIsValid(typeTuple))
-		return (int) ((Form_pg_type) GETSTRUCT(typeTuple))->typoutput;
-
-	elog(ERROR, "GetOutputFunction: Cache lookup of type %u failed", type);
-	return InvalidOid;
+	typeTuple = SearchSysCache(TYPEOID,
+							   ObjectIdGetDatum(type),
+							   0, 0, 0);
+	if (!HeapTupleIsValid(typeTuple))
+		elog(ERROR, "GetInputFunction: Cache lookup of type %u failed", type);
+	result = ((Form_pg_type) GETSTRUCT(typeTuple))->typinput;
+	ReleaseSysCache(typeTuple);
+	return result;
 }
 
 static Oid
 GetTypeElement(Oid type)
 {
 	HeapTuple	typeTuple;
+	Oid			result;
 
-	typeTuple = SearchSysCacheTuple(TYPEOID,
-									ObjectIdGetDatum(type),
-									0, 0, 0);
-
-	if (HeapTupleIsValid(typeTuple))
-		return (int) ((Form_pg_type) GETSTRUCT(typeTuple))->typelem;
-
-	elog(ERROR, "GetOutputFunction: Cache lookup of type %u failed", type);
-	return InvalidOid;
+	typeTuple = SearchSysCache(TYPEOID,
+							   ObjectIdGetDatum(type),
+							   0, 0, 0);
+	if (!HeapTupleIsValid(typeTuple))
+		elog(ERROR, "GetTypeElement: Cache lookup of type %u failed", type);
+	result = ((Form_pg_type) GETSTRUCT(typeTuple))->typelem;
+	ReleaseSysCache(typeTuple);
+	return result;
 }
 
-static Oid
-GetInputFunction(Oid type)
-{
-	HeapTuple	typeTuple;
-
-	typeTuple = SearchSysCacheTuple(TYPEOID,
-									ObjectIdGetDatum(type),
-									0, 0, 0);
-
-	if (HeapTupleIsValid(typeTuple))
-		return (int) ((Form_pg_type) GETSTRUCT(typeTuple))->typinput;
-
-	elog(ERROR, "GetInputFunction: Cache lookup of type %u failed", type);
-	return InvalidOid;
-}
-
-static Oid
-IsTypeByVal(Oid type)
-{
-	HeapTuple	typeTuple;
-
-	typeTuple = SearchSysCacheTuple(TYPEOID,
-									ObjectIdGetDatum(type),
-									0, 0, 0);
-
-	if (HeapTupleIsValid(typeTuple))
-		return (int) ((Form_pg_type) GETSTRUCT(typeTuple))->typbyval;
-
-	elog(ERROR, "GetInputFunction: Cache lookup of type %u failed", type);
-
-	return InvalidOid;
-}
-
-/*
- * Given the OID of a relation, return an array of index relation descriptors
- * and the number of index relations.  These relation descriptors are open
- * using index_open().
- *
- * Space for the array itself is palloc'ed.
- */
-
-typedef struct rel_list
-{
-	Oid			index_rel_oid;
-	struct rel_list *next;
-} RelationList;
-
-static void
-GetIndexRelations(Oid main_relation_oid,
-				  int *n_indices,
-				  Relation **index_rels)
-{
-	RelationList *head,
-			   *scan;
-	Relation	pg_index_rel;
-	HeapScanDesc scandesc;
-	Oid			index_relation_oid;
-	HeapTuple	tuple;
-	TupleDesc	tupDesc;
-	int			i;
-	bool		isnull;
-
-	pg_index_rel = heap_openr(IndexRelationName, AccessShareLock);
-	scandesc = heap_beginscan(pg_index_rel, 0, SnapshotNow, 0, NULL);
-	tupDesc = RelationGetDescr(pg_index_rel);
-
-	*n_indices = 0;
-
-	head = (RelationList *) palloc(sizeof(RelationList));
-	scan = head;
-	head->next = NULL;
-
-	while (HeapTupleIsValid(tuple = heap_getnext(scandesc, 0)))
-	{
-
-		index_relation_oid = (Oid) DatumGetInt32(heap_getattr(tuple, 2,
-													  tupDesc, &isnull));
-		if (index_relation_oid == main_relation_oid)
-		{
-			scan->index_rel_oid = (Oid) DatumGetInt32(heap_getattr(tuple,
-												Anum_pg_index_indexrelid,
-													  tupDesc, &isnull));
-			(*n_indices)++;
-			scan->next = (RelationList *) palloc(sizeof(RelationList));
-			scan = scan->next;
-		}
-	}
-
-	heap_endscan(scandesc);
-	heap_close(pg_index_rel, AccessShareLock);
-
-	/* We cannot trust to relhasindex of the main_relation now, so... */
-	if (*n_indices == 0)
-		return;
-
-	*index_rels = (Relation *) palloc(*n_indices * sizeof(Relation));
-
-	for (i = 0, scan = head; i < *n_indices; i++, scan = scan->next)
-	{
-		(*index_rels)[i] = index_open(scan->index_rel_oid);
-		/* see comments in ExecOpenIndices() in execUtils.c */
-		if ((*index_rels)[i] != NULL &&
-			((*index_rels)[i])->rd_rel->relam != BTREE_AM_OID &&
-			((*index_rels)[i])->rd_rel->relam != HASH_AM_OID)
-			LockRelation((*index_rels)[i], AccessExclusiveLock);
-	}
-
-	for (i = 0, scan = head; i < *n_indices + 1; i++)
-	{
-		scan = head->next;
-		pfree(head);
-		head = scan;
-	}
-}
 
 /*
  * Reads input from fp until an end of line is seen.
@@ -1315,29 +1119,35 @@ CopyReadAttribute(FILE *fp, bool *isnull, char *delim, int *newline, char *null_
 		}
 		appendStringInfoCharMacro(&attribute_buf, c);
 #ifdef MULTIBYTE
-		/* get additional bytes of the char, if any */
-		s[0] = c;
-		mblen = pg_encoding_mblen(encoding, s);
-		for (j = 1; j < mblen; j++)
+		if (client_encoding != server_encoding)
 		{
-			c = CopyGetChar(fp);
-			if (c == EOF)
-				goto endOfFile;
-			appendStringInfoCharMacro(&attribute_buf, c);
+			/* get additional bytes of the char, if any */
+			s[0] = c;
+			mblen = pg_encoding_mblen(client_encoding, s);
+			for (j = 1; j < mblen; j++)
+			{
+				c = CopyGetChar(fp);
+				if (c == EOF)
+					goto endOfFile;
+				appendStringInfoCharMacro(&attribute_buf, c);
+			}
 		}
 #endif
 	}
 
 #ifdef MULTIBYTE
-	cvt = (char *) pg_client_to_server((unsigned char *) attribute_buf.data,
-									   attribute_buf.len);
-	if (cvt != attribute_buf.data)
+	if (client_encoding != server_encoding)
 	{
-		/* transfer converted data back to attribute_buf */
-		attribute_buf.len = 0;
-		attribute_buf.data[0] = '\0';
-		appendBinaryStringInfo(&attribute_buf, cvt, strlen(cvt));
-		pfree(cvt);
+		cvt = (char *) pg_client_to_server((unsigned char *) attribute_buf.data,
+										   attribute_buf.len);
+		if (cvt != attribute_buf.data)
+		{
+			/* transfer converted data back to attribute_buf */
+			attribute_buf.len = 0;
+			attribute_buf.data[0] = '\0';
+			appendBinaryStringInfo(&attribute_buf, cvt, strlen(cvt));
+			pfree(cvt);
+		}
 	}
 #endif
 
@@ -1364,15 +1174,23 @@ CopyAttributeOut(FILE *fp, char *server_string, char *delim)
 #endif
 
 #ifdef MULTIBYTE
-	string = (char *) pg_server_to_client((unsigned char *) server_string,
-										  strlen(server_string));
-	string_start = string;
+	if (client_encoding != server_encoding)
+	{
+		string = (char *) pg_server_to_client((unsigned char *) server_string,
+											  strlen(server_string));
+		string_start = string;
+	}
+	else
+	{
+		string = server_string;
+		string_start = NULL;	/* unused, but keep compiler quiet */
+	}
 #else
 	string = server_string;
 #endif
 
 #ifdef MULTIBYTE
-	for (; (mblen = pg_encoding_mblen(encoding, string)) &&
+	for (; (mblen = (server_encoding == client_encoding ? 1 : pg_encoding_mblen(client_encoding, string))) &&
 		 ((c = *string) != '\0'); string += mblen)
 #else
 	for (; (c = *string) != '\0'; string++)
@@ -1389,30 +1207,7 @@ CopyAttributeOut(FILE *fp, char *server_string, char *delim)
 	}
 
 #ifdef MULTIBYTE
-	if (string_start != server_string)
+	if (client_encoding != server_encoding)
 		pfree(string_start);	/* pfree pg_server_to_client result */
 #endif
 }
-
-/*
- * Returns the number of tuples in a relation.	Unfortunately, currently
- * must do a scan of the entire relation to determine this.
- *
- * relation is expected to be an open relation descriptor.
- */
-static int
-CountTuples(Relation relation)
-{
-	HeapScanDesc scandesc;
-	HeapTuple	tuple;
-
-	int			i;
-
-	scandesc = heap_beginscan(relation, 0, QuerySnapshot, 0, NULL);
-
-	i = 0;
-	while (HeapTupleIsValid(tuple = heap_getnext(scandesc, 0)))
-		i++;
-	heap_endscan(scandesc);
-	return i;
-}
