@@ -3,12 +3,12 @@
  * nodeUnique.c
  *	  Routines to handle unique'ing of queries where appropriate
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/executor/nodeUnique.c,v 1.28 2000/04/12 17:15:10 momjian Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/executor/nodeUnique.c,v 1.32 2001/03/22 06:16:13 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -27,7 +27,6 @@
 #include "postgres.h"
 
 #include "access/heapam.h"
-#include "access/printtup.h"
 #include "executor/executor.h"
 #include "executor/nodeGroup.h"
 #include "executor/nodeUnique.h"
@@ -48,68 +47,67 @@ ExecUnique(Unique *node)
 	Plan	   *outerPlan;
 	TupleDesc	tupDesc;
 
-	/* ----------------
-	 *	get information from the node
-	 * ----------------
+	/*
+	 * get information from the node
 	 */
 	uniquestate = node->uniquestate;
 	outerPlan = outerPlan((Plan *) node);
 	resultTupleSlot = uniquestate->cstate.cs_ResultTupleSlot;
 	tupDesc = ExecGetResultType(&uniquestate->cstate);
 
-	/* ----------------
-	 *	now loop, returning only non-duplicate tuples.
-	 *	We assume that the tuples arrive in sorted order
-	 *	so we can detect duplicates easily.
-	 * ----------------
+	/*
+	 * now loop, returning only non-duplicate tuples. We assume that the
+	 * tuples arrive in sorted order so we can detect duplicates easily.
 	 */
 	for (;;)
 	{
-		/* ----------------
-		 *	 fetch a tuple from the outer subplan
-		 * ----------------
+
+		/*
+		 * fetch a tuple from the outer subplan
 		 */
 		slot = ExecProcNode(outerPlan, (Plan *) node);
 		if (TupIsNull(slot))
 			return NULL;
 
-		/* ----------------
-		 *	 Always return the first tuple from the subplan.
-		 * ----------------
+		/*
+		 * Always return the first tuple from the subplan.
 		 */
 		if (uniquestate->priorTuple == NULL)
 			break;
 
-		/* ----------------
-		 *	 Else test if the new tuple and the previously returned
-		 *	 tuple match.  If so then we loop back and fetch
-		 *	 another new tuple from the subplan.
-		 * ----------------
+		/*
+		 * Else test if the new tuple and the previously returned tuple
+		 * match.  If so then we loop back and fetch another new tuple
+		 * from the subplan.
 		 */
 		if (!execTuplesMatch(slot->val, uniquestate->priorTuple,
 							 tupDesc,
 							 node->numCols, node->uniqColIdx,
-							 uniquestate->eqfunctions))
+							 uniquestate->eqfunctions,
+							 uniquestate->tempContext))
 			break;
 	}
 
-	/* ----------------
-	 *	We have a new tuple different from the previous saved tuple (if any).
-	 *	Save it and return it.	Note that we make two copies of the tuple:
-	 *	one to keep for our own future comparisons, and one to return to the
-	 *	caller.  We need to copy the tuple returned by the subplan to avoid
-	 *	holding buffer refcounts, and we need our own copy because the caller
-	 *	may alter the resultTupleSlot (eg via ExecRemoveJunk).
-	 * ----------------
+	/*
+	 * We have a new tuple different from the previous saved tuple (if
+	 * any). Save it and return it.  We must copy it because the source
+	 * subplan won't guarantee that this source tuple is still accessible
+	 * after fetching the next source tuple.
+	 *
+	 * Note that we manage the copy ourselves.	We can't rely on the result
+	 * tuple slot to maintain the tuple reference because our caller may
+	 * replace the slot contents with a different tuple (see junk filter
+	 * handling in execMain.c).  We assume that the caller will no longer
+	 * be interested in the current tuple after he next calls us.
 	 */
 	if (uniquestate->priorTuple != NULL)
 		heap_freetuple(uniquestate->priorTuple);
 	uniquestate->priorTuple = heap_copytuple(slot->val);
 
-	ExecStoreTuple(heap_copytuple(slot->val),
+	ExecStoreTuple(uniquestate->priorTuple,
 				   resultTupleSlot,
 				   InvalidBuffer,
-				   true);
+				   false);		/* tuple does not belong to slot */
 
 	return resultTupleSlot;
 }
@@ -127,50 +125,48 @@ ExecInitUnique(Unique *node, EState *estate, Plan *parent)
 	UniqueState *uniquestate;
 	Plan	   *outerPlan;
 
-	/* ----------------
-	 *	assign execution state to node
-	 * ----------------
+	/*
+	 * assign execution state to node
 	 */
 	node->plan.state = estate;
 
-	/* ----------------
-	 *	create new UniqueState for node
-	 * ----------------
+	/*
+	 * create new UniqueState for node
 	 */
 	uniquestate = makeNode(UniqueState);
 	node->uniquestate = uniquestate;
 	uniquestate->priorTuple = NULL;
 
-	/* ----------------
-	 *	Miscellaneous initialization
+	/*
+	 * Miscellaneous initialization
 	 *
-	 *		 +	assign node's base_id
-	 *		 +	assign debugging hooks and
-	 *
-	 *	Unique nodes have no ExprContext initialization because
-	 *	they never call ExecQual or ExecTargetList.
-	 * ----------------
+	 * Unique nodes have no ExprContext initialization because they never
+	 * call ExecQual or ExecProject.  But they do need a per-tuple memory
+	 * context anyway for calling execTuplesMatch.
 	 */
-	ExecAssignNodeBaseInfo(estate, &uniquestate->cstate, parent);
+	uniquestate->tempContext =
+		AllocSetContextCreate(CurrentMemoryContext,
+							  "Unique",
+							  ALLOCSET_DEFAULT_MINSIZE,
+							  ALLOCSET_DEFAULT_INITSIZE,
+							  ALLOCSET_DEFAULT_MAXSIZE);
 
 #define UNIQUE_NSLOTS 1
-	/* ------------
+
+	/*
 	 * Tuple table initialization
-	 * ------------
 	 */
 	ExecInitResultTupleSlot(estate, &uniquestate->cstate);
 
-	/* ----------------
-	 *	then initialize outer plan
-	 * ----------------
+	/*
+	 * then initialize outer plan
 	 */
 	outerPlan = outerPlan((Plan *) node);
 	ExecInitNode(outerPlan, estate, (Plan *) node);
 
-	/* ----------------
-	 *	unique nodes do no projections, so initialize
-	 *	projection info for this node appropriately
-	 * ----------------
+	/*
+	 * unique nodes do no projections, so initialize projection info for
+	 * this node appropriately
 	 */
 	ExecAssignResultTypeFromOuterPlan((Plan *) node, &uniquestate->cstate);
 	uniquestate->cstate.cs_ProjInfo = NULL;
@@ -208,6 +204,8 @@ ExecEndUnique(Unique *node)
 
 	ExecEndNode(outerPlan((Plan *) node), (Plan *) node);
 
+	MemoryContextDelete(uniquestate->tempContext);
+
 	/* clean up tuple table */
 	ExecClearTuple(uniquestate->cstate.cs_ResultTupleSlot);
 	if (uniquestate->priorTuple != NULL)
