@@ -3,12 +3,12 @@
  * parse_expr.c
  *	  handle expressions in parser
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/parser/parse_expr.c,v 1.76 2000/04/12 17:15:26 momjian Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/parser/parse_expr.c,v 1.92 2001/03/22 03:59:41 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -19,16 +19,16 @@
 #include "catalog/pg_proc.h"
 #include "nodes/makefuncs.h"
 #include "nodes/params.h"
-#include "nodes/relation.h"
-#include "parse.h"
 #include "parser/analyze.h"
 #include "parser/gramparse.h"
+#include "parser/parse.h"
 #include "parser/parse_coerce.h"
 #include "parser/parse_expr.h"
 #include "parser/parse_func.h"
 #include "parser/parse_oper.h"
 #include "parser/parse_relation.h"
 #include "parser/parse_target.h"
+#include "parser/parse_type.h"
 #include "utils/builtins.h"
 #include "utils/syscache.h"
 
@@ -133,7 +133,6 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 				param->paramid = (AttrNumber) paramno;
 				param->paramname = "<unnamed>";
 				param->paramtype = toid;
-				param->param_tlist = NIL;
 				result = transformIndirection(pstate, (Node *) param,
 											  pno->indirection);
 				/* cope with typecast applied to param */
@@ -158,41 +157,51 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 				{
 					case OP:
 						{
-							Node	   *lexpr = transformExpr(pstate, a->lexpr, precedence);
-							Node	   *rexpr = transformExpr(pstate, a->rexpr, precedence);
+							Node	   *lexpr = transformExpr(pstate,
+															  a->lexpr,
+															  precedence);
+							Node	   *rexpr = transformExpr(pstate,
+															  a->rexpr,
+															  precedence);
 
 							result = (Node *) make_op(a->opname, lexpr, rexpr);
 						}
 						break;
 					case ISNULL:
 						{
-							Node	   *lexpr = transformExpr(pstate, a->lexpr, precedence);
+							Node	   *lexpr = transformExpr(pstate,
+															  a->lexpr,
+															  precedence);
 
 							result = ParseFuncOrColumn(pstate,
 													   "nullvalue",
-													   lcons(lexpr, NIL),
+													   makeList1(lexpr),
 													   false, false,
-												   &pstate->p_last_resno,
 													   precedence);
 						}
 						break;
 					case NOTNULL:
 						{
-							Node	   *lexpr = transformExpr(pstate, a->lexpr, precedence);
+							Node	   *lexpr = transformExpr(pstate,
+															  a->lexpr,
+															  precedence);
 
 							result = ParseFuncOrColumn(pstate,
 													   "nonnullvalue",
-													   lcons(lexpr, NIL),
+													   makeList1(lexpr),
 													   false, false,
-												   &pstate->p_last_resno,
 													   precedence);
 						}
 						break;
 					case AND:
 						{
+							Node	   *lexpr = transformExpr(pstate,
+															  a->lexpr,
+															  precedence);
+							Node	   *rexpr = transformExpr(pstate,
+															  a->rexpr,
+															  precedence);
 							Expr	   *expr = makeNode(Expr);
-							Node	   *lexpr = transformExpr(pstate, a->lexpr, precedence);
-							Node	   *rexpr = transformExpr(pstate, a->rexpr, precedence);
 
 							if (exprType(lexpr) != BOOLOID)
 								elog(ERROR, "left-hand side of AND is type '%s', not '%s'",
@@ -204,15 +213,19 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 
 							expr->typeOid = BOOLOID;
 							expr->opType = AND_EXPR;
-							expr->args = makeList(lexpr, rexpr, -1);
+							expr->args = makeList2(lexpr, rexpr);
 							result = (Node *) expr;
 						}
 						break;
 					case OR:
 						{
+							Node	   *lexpr = transformExpr(pstate,
+															  a->lexpr,
+															  precedence);
+							Node	   *rexpr = transformExpr(pstate,
+															  a->rexpr,
+															  precedence);
 							Expr	   *expr = makeNode(Expr);
-							Node	   *lexpr = transformExpr(pstate, a->lexpr, precedence);
-							Node	   *rexpr = transformExpr(pstate, a->rexpr, precedence);
 
 							if (exprType(lexpr) != BOOLOID)
 								elog(ERROR, "left-hand side of OR is type '%s', not '%s'",
@@ -222,21 +235,23 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 									 typeidTypeName(exprType(rexpr)), typeidTypeName(BOOLOID));
 							expr->typeOid = BOOLOID;
 							expr->opType = OR_EXPR;
-							expr->args = makeList(lexpr, rexpr, -1);
+							expr->args = makeList2(lexpr, rexpr);
 							result = (Node *) expr;
 						}
 						break;
 					case NOT:
 						{
+							Node	   *rexpr = transformExpr(pstate,
+															  a->rexpr,
+															  precedence);
 							Expr	   *expr = makeNode(Expr);
-							Node	   *rexpr = transformExpr(pstate, a->rexpr, precedence);
 
 							if (exprType(rexpr) != BOOLOID)
 								elog(ERROR, "argument to NOT is type '%s', not '%s'",
 									 typeidTypeName(exprType(rexpr)), typeidTypeName(BOOLOID));
 							expr->typeOid = BOOLOID;
 							expr->opType = NOT_EXPR;
-							expr->args = makeList(rexpr, -1);
+							expr->args = makeList1(rexpr);
 							result = (Node *) expr;
 						}
 						break;
@@ -255,13 +270,14 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 
 				/* transform the list of arguments */
 				foreach(args, fn->args)
-					lfirst(args) = transformExpr(pstate, (Node *) lfirst(args), precedence);
+					lfirst(args) = transformExpr(pstate,
+												 (Node *) lfirst(args),
+												 precedence);
 				result = ParseFuncOrColumn(pstate,
 										   fn->funcname,
 										   fn->args,
 										   fn->agg_star,
 										   fn->agg_distinct,
-										   &pstate->p_last_resno,
 										   precedence);
 				break;
 			}
@@ -278,7 +294,7 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 					break;
 				}
 				pstate->p_hasSubLinks = true;
-				qtrees = parse_analyze(lcons(sublink->subselect, NIL), pstate);
+				qtrees = parse_analyze(sublink->subselect, pstate);
 				if (length(qtrees) != 1)
 					elog(ERROR, "Bad query in subselect");
 				qtree = (Query *) lfirst(qtrees);
@@ -324,23 +340,25 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 				else
 				{
 					/* ALL, ANY, or MULTIEXPR: generate operator list */
-					char	   *op = lfirst(sublink->oper);
 					List	   *left_list = sublink->lefthand;
 					List	   *right_list = qtree->targetList;
+					char	   *op;
 					List	   *elist;
 
 					foreach(elist, left_list)
 						lfirst(elist) = transformExpr(pstate, lfirst(elist),
 													  precedence);
 
+					Assert(IsA(sublink->oper, A_Expr));
+					op = ((A_Expr *) sublink->oper)->opname;
+					sublink->oper = NIL;
+
 					/* Combining operators other than =/<> is dubious... */
 					if (length(left_list) != 1 &&
 						strcmp(op, "=") != 0 && strcmp(op, "<>") != 0)
 						elog(ERROR, "Row comparison cannot use '%s'",
 							 op);
 
-					sublink->oper = NIL;
-
 					/*
 					 * Scan subquery's targetlist to find values that will
 					 * be matched against lefthand values.	We need to
@@ -365,10 +383,15 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 						lexpr = lfirst(left_list);
 						left_list = lnext(left_list);
 
+						/*
+						 * It's OK to use oper() not compatible_oper()
+						 * here, because make_subplan() will insert type
+						 * coercion calls if needed.
+						 */
 						optup = oper(op,
 									 exprType(lexpr),
 									 exprType(tent->expr),
-									 FALSE);
+									 false);
 						opform = (Form_pg_operator) GETSTRUCT(optup);
 
 						if (opform->oprresult != BOOLOID)
@@ -379,10 +402,9 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 
 						newop = makeOper(oprid(optup),	/* opno */
 										 InvalidOid,	/* opid */
-										 opform->oprresult,
-										 0,
-										 NULL);
+										 opform->oprresult);
 						sublink->oper = lappend(sublink->oper, newop);
+						ReleaseSysCache(optup);
 					}
 					if (left_list != NIL)
 						elog(ERROR, "Subselect has too few fields");
@@ -395,9 +417,9 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 			{
 				CaseExpr   *c = (CaseExpr *) expr;
 				CaseWhen   *w;
+				List	   *typeids = NIL;
 				List	   *args;
 				Oid			ptype;
-				CATEGORY	pcategory;
 
 				/* transform the list of arguments */
 				foreach(args, c->args)
@@ -415,6 +437,7 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 						w->expr = (Node *) a;
 					}
 					lfirst(args) = transformExpr(pstate, (Node *) w, precedence);
+					typeids = lappendi(typeids, exprType(w->result));
 				}
 
 				/*
@@ -436,103 +459,27 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 				}
 				c->defresult = transformExpr(pstate, c->defresult, precedence);
 
-				/* now check types across result clauses... */
-				c->casetype = exprType(c->defresult);
-				ptype = c->casetype;
-				pcategory = TypeCategory(ptype);
-				foreach(args, c->args)
-				{
-					Oid			wtype;
+				/*
+				 * Note: default result is considered the most significant
+				 * type in determining preferred type.	This is how the
+				 * code worked before, but it seems a little bogus to me
+				 * --- tgl
+				 */
+				typeids = lconsi(exprType(c->defresult), typeids);
 
-					w = lfirst(args);
-					wtype = exprType(w->result);
-					/* move on to next one if no new information... */
-					if (wtype && (wtype != UNKNOWNOID)
-						&& (wtype != ptype))
-					{
-						if (!ptype || ptype == UNKNOWNOID)
-						{
-							/* so far, only nulls so take anything... */
-							ptype = wtype;
-							pcategory = TypeCategory(ptype);
-						}
-						else if ((TypeCategory(wtype) != pcategory)
-								 || ((TypeCategory(wtype) == USER_TYPE)
-							&& (TypeCategory(c->casetype) == USER_TYPE)))
-						{
-
-							/*
-							 * both types in different categories? then
-							 * not much hope...
-							 */
-							elog(ERROR, "CASE/WHEN types '%s' and '%s' not matched",
-								 typeidTypeName(c->casetype), typeidTypeName(wtype));
-						}
-						else if (IsPreferredType(pcategory, wtype)
-								 && can_coerce_type(1, &ptype, &wtype))
-						{
-
-							/*
-							 * new one is preferred and can convert? then
-							 * take it...
-							 */
-							ptype = wtype;
-							pcategory = TypeCategory(ptype);
-						}
-					}
-				}
+				ptype = select_common_type(typeids, "CASE");
+				c->casetype = ptype;
 
 				/* Convert default result clause, if necessary */
-				if (c->casetype != ptype)
-				{
-					if (!c->casetype || c->casetype == UNKNOWNOID)
-					{
+				c->defresult = coerce_to_common_type(pstate, c->defresult,
+													 ptype, "CASE/ELSE");
 
-						/*
-						 * default clause is NULL, so assign preferred
-						 * type from WHEN clauses...
-						 */
-						c->casetype = ptype;
-					}
-					else if (can_coerce_type(1, &c->casetype, &ptype))
-					{
-						c->defresult = coerce_type(pstate, c->defresult,
-												 c->casetype, ptype, -1);
-						c->casetype = ptype;
-					}
-					else
-					{
-						elog(ERROR, "CASE/ELSE unable to convert to type '%s'",
-							 typeidTypeName(ptype));
-					}
-				}
-
-				/* Convert when clauses, if not null and if necessary */
+				/* Convert when-clause results, if necessary */
 				foreach(args, c->args)
 				{
-					Oid			wtype;
-
 					w = lfirst(args);
-					wtype = exprType(w->result);
-
-					/*
-					 * only bother with conversion if not NULL and
-					 * different type...
-					 */
-					if (wtype && (wtype != UNKNOWNOID)
-						&& (wtype != ptype))
-					{
-						if (can_coerce_type(1, &wtype, &ptype))
-						{
-							w->result = coerce_type(pstate, w->result, wtype,
-													ptype, -1);
-						}
-						else
-						{
-							elog(ERROR, "CASE/WHEN unable to convert to type '%s'",
-								 typeidTypeName(ptype));
-						}
-					}
+					w->result = coerce_to_common_type(pstate, w->result,
+													  ptype, "CASE/WHEN");
 				}
 
 				result = expr;
@@ -543,7 +490,7 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 			{
 				CaseWhen   *w = (CaseWhen *) expr;
 
-				w->expr = transformExpr(pstate, (Node *) w->expr, precedence);
+				w->expr = transformExpr(pstate, w->expr, precedence);
 				if (exprType(w->expr) != BOOLOID)
 					elog(ERROR, "WHEN clause must have a boolean result");
 
@@ -558,7 +505,7 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 					n->val.type = T_Null;
 					w->result = (Node *) n;
 				}
-				w->result = transformExpr(pstate, (Node *) w->result, precedence);
+				w->result = transformExpr(pstate, w->result, precedence);
 				result = expr;
 				break;
 			}
@@ -577,6 +524,7 @@ transformExpr(ParseState *pstate, Node *expr, int precedence)
 		case T_Param:
 		case T_Aggref:
 		case T_ArrayRef:
+		case T_FieldSelect:
 		case T_RelabelType:
 			{
 				result = (Node *) expr;
@@ -600,7 +548,8 @@ transformIndirection(ParseState *pstate, Node *basenode, List *indirection)
 {
 	if (indirection == NIL)
 		return basenode;
-	return (Node *) transformArraySubscripts(pstate, basenode,
+	return (Node *) transformArraySubscripts(pstate,
+											 basenode, exprType(basenode),
 											 indirection, false, NULL);
 }
 
@@ -609,8 +558,7 @@ transformAttr(ParseState *pstate, Attr *att, int precedence)
 {
 	Node	   *basenode;
 
-	basenode = ParseNestedFuncOrColumn(pstate, att, &pstate->p_last_resno,
-									   precedence);
+	basenode = ParseNestedFuncOrColumn(pstate, att, precedence);
 	return transformIndirection(pstate, basenode, att->indirection);
 }
 
@@ -618,14 +566,14 @@ static Node *
 transformIdent(ParseState *pstate, Ident *ident, int precedence)
 {
 	Node	   *result = NULL;
-	RangeTblEntry *rte;
+	int			sublevels_up;
 
 	/*
 	 * try to find the ident as a relation ... but not if subscripts
 	 * appear
 	 */
 	if (ident->indirection == NIL &&
-		refnameRangeTableEntry(pstate, ident->name) != NULL)
+	 refnameRangeOrJoinEntry(pstate, ident->name, &sublevels_up) != NULL)
 	{
 		ident->isRel = TRUE;
 		result = (Node *) ident;
@@ -634,14 +582,10 @@ transformIdent(ParseState *pstate, Ident *ident, int precedence)
 	if (result == NULL || precedence == EXPR_COLUMN_FIRST)
 	{
 		/* try to find the ident as a column */
-		if ((rte = colnameRangeTableEntry(pstate, ident->name)) != NULL)
-		{
-			/* Convert it to a fully qualified Attr, and transform that */
-			Attr	   *att = makeAttr(rte->eref->relname, ident->name);
+		Node	   *var = colnameToVar(pstate, ident->name);
 
-			att->indirection = ident->indirection;
-			return transformAttr(pstate, att, precedence);
-		}
+		if (var != NULL)
+			result = transformIndirection(pstate, var, ident->indirection);
 	}
 
 	if (result == NULL)
@@ -688,6 +632,9 @@ exprType(Node *expr)
 		case T_Param:
 			type = ((Param *) expr)->paramtype;
 			break;
+		case T_FieldSelect:
+			type = ((FieldSelect *) expr)->resulttype;
+			break;
 		case T_RelabelType:
 			type = ((RelabelType *) expr)->resulttype;
 			break;
@@ -771,6 +718,9 @@ exprTypmod(Node *expr)
 					return coercedTypmod;
 			}
 			break;
+		case T_FieldSelect:
+			return ((FieldSelect *) expr)->resulttypmod;
+			break;
 		case T_RelabelType:
 			return ((RelabelType *) expr)->resulttypmod;
 			break;
@@ -800,7 +750,8 @@ exprIsLengthCoercion(Node *expr, int32 *coercedTypmod)
 {
 	Func	   *func;
 	Const	   *second_arg;
-	HeapTuple	tup;
+	HeapTuple	procTuple;
+	HeapTuple	typeTuple;
 	Form_pg_proc procStruct;
 	Form_pg_type typeStruct;
 
@@ -831,12 +782,12 @@ exprIsLengthCoercion(Node *expr, int32 *coercedTypmod)
 	/*
 	 * Lookup the function in pg_proc
 	 */
-	tup = SearchSysCacheTuple(PROCOID,
-							  ObjectIdGetDatum(func->funcid),
-							  0, 0, 0);
-	if (!HeapTupleIsValid(tup))
+	procTuple = SearchSysCache(PROCOID,
+							   ObjectIdGetDatum(func->funcid),
+							   0, 0, 0);
+	if (!HeapTupleIsValid(procTuple))
 		elog(ERROR, "cache lookup for proc %u failed", func->funcid);
-	procStruct = (Form_pg_proc) GETSTRUCT(tup);
+	procStruct = (Form_pg_proc) GETSTRUCT(procTuple);
 
 	/*
 	 * It must be a function with two arguments where the first is of the
@@ -847,29 +798,39 @@ exprIsLengthCoercion(Node *expr, int32 *coercedTypmod)
 		procStruct->prorettype != procStruct->proargtypes[0] ||
 		procStruct->proargtypes[1] != INT4OID ||
 		procStruct->prorettype != ((Expr *) expr)->typeOid)
+	{
+		ReleaseSysCache(procTuple);
 		return false;
+	}
 
 	/*
 	 * Furthermore, the name of the function must be the same as the
 	 * argument/result type's name.
 	 */
-	tup = SearchSysCacheTuple(TYPEOID,
-							  ObjectIdGetDatum(procStruct->prorettype),
-							  0, 0, 0);
-	if (!HeapTupleIsValid(tup))
+	typeTuple = SearchSysCache(TYPEOID,
+							   ObjectIdGetDatum(procStruct->prorettype),
+							   0, 0, 0);
+	if (!HeapTupleIsValid(typeTuple))
 		elog(ERROR, "cache lookup for type %u failed",
 			 procStruct->prorettype);
-	typeStruct = (Form_pg_type) GETSTRUCT(tup);
+	typeStruct = (Form_pg_type) GETSTRUCT(typeTuple);
 	if (strncmp(NameStr(procStruct->proname),
 				NameStr(typeStruct->typname),
 				NAMEDATALEN) != 0)
+	{
+		ReleaseSysCache(procTuple);
+		ReleaseSysCache(typeTuple);
 		return false;
+	}
 
 	/*
 	 * OK, it is indeed a length-coercion function.
 	 */
 	if (coercedTypmod != NULL)
 		*coercedTypmod = DatumGetInt32(second_arg->constvalue);
+
+	ReleaseSysCache(procTuple);
+	ReleaseSysCache(typeTuple);
 	return true;
 }
 
@@ -880,21 +841,25 @@ exprIsLengthCoercion(Node *expr, int32 *coercedTypmod)
 static Node *
 parser_typecast_constant(Value *expr, TypeName *typename)
 {
-	Const	   *con;
 	Type		tp;
 	Datum		datum;
+	Const	   *con;
 	char	   *const_string = NULL;
 	bool		string_palloced = false;
 	bool		isNull = false;
 
+	tp = typenameType(TypeNameToInternalName(typename));
+
 	switch (nodeTag(expr))
 	{
 		case T_Integer:
+			const_string = DatumGetCString(DirectFunctionCall1(int4out,
+										 Int32GetDatum(expr->val.ival)));
 			string_palloced = true;
-			const_string = int4out(expr->val.ival);
 			break;
 		case T_Float:
 		case T_String:
+		case T_BitString:
 			const_string = expr->val.str;
 			break;
 		case T_Null:
@@ -902,19 +867,9 @@ parser_typecast_constant(Value *expr, TypeName *typename)
 			break;
 		default:
 			elog(ERROR, "Cannot cast this expression to type '%s'",
-				 typename->name);
+				 typeTypeName(tp));
 	}
 
-	if (typename->arrayBounds != NIL)
-	{
-		char		type_string[NAMEDATALEN + 2];
-
-		sprintf(type_string, "_%s", typename->name);
-		tp = (Type) typenameType(type_string);
-	}
-	else
-		tp = (Type) typenameType(typename->name);
-
 	if (isNull)
 		datum = (Datum) NULL;
 	else
@@ -931,6 +886,8 @@ parser_typecast_constant(Value *expr, TypeName *typename)
 	if (string_palloced)
 		pfree(const_string);
 
+	ReleaseSysCache(tp);
+
 	return (Node *) con;
 }
 
@@ -947,19 +904,9 @@ parser_typecast_expression(ParseState *pstate,
 						   Node *expr, TypeName *typename)
 {
 	Oid			inputType = exprType(expr);
-	Type		tp;
 	Oid			targetType;
 
-	if (typename->arrayBounds != NIL)
-	{
-		char		type_string[NAMEDATALEN + 2];
-
-		sprintf(type_string, "_%s", typename->name);
-		tp = (Type) typenameType(type_string);
-	}
-	else
-		tp = (Type) typenameType(typename->name);
-	targetType = typeTypeId(tp);
+	targetType = typenameTypeId(TypeNameToInternalName(typename));
 
 	if (inputType == InvalidOid)
 		return expr;			/* do nothing if NULL input */
@@ -983,3 +930,27 @@ parser_typecast_expression(ParseState *pstate,
 
 	return expr;
 }
+
+/*
+ * Given a TypeName node as returned by the grammar, generate the internal
+ * name of the corresponding type.	Note this does NOT check if the type
+ * exists or not.
+ */
+char *
+TypeNameToInternalName(TypeName *typename)
+{
+	if (typename->arrayBounds != NIL)
+	{
+
+		/*
+		 * By convention, the name of an array type is the name of its
+		 * element type with "_" prepended.
+		 */
+		char	   *arrayname = palloc(strlen(typename->name) + 2);
+
+		sprintf(arrayname, "_%s", typename->name);
+		return arrayname;
+	}
+	else
+		return typename->name;
+}
