@@ -5,12 +5,12 @@
  *	  commands.  At one time acted as an interface between the Lisp and C
  *	  systems.
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /usr/local/cvsroot/pgsql/src/backend/tcop/utility.c,v 1.87 2000/04/25 10:38:38 inoue Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/tcop/utility.c,v 1.109 2001/03/22 06:16:17 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,7 +18,7 @@
 
 #include "access/heapam.h"
 #include "catalog/catalog.h"
-#include "catalog/pg_type.h"
+#include "catalog/pg_shadow.h"
 #include "commands/async.h"
 #include "commands/cluster.h"
 #include "commands/command.h"
@@ -38,31 +38,95 @@
 #include "commands/view.h"
 #include "miscadmin.h"
 #include "parser/parse.h"
+#include "parser/parse_clause.h"
+#include "parser/parse_expr.h"
 #include "rewrite/rewriteDefine.h"
 #include "rewrite/rewriteRemove.h"
 #include "tcop/utility.h"
 #include "utils/acl.h"
 #include "utils/ps_status.h"
 #include "utils/syscache.h"
+#include "access/xlog.h"
 
-
-/* ----------------
- *		CHECK_IF_ABORTED() is used to avoid doing unnecessary
- *		processing within an aborted transaction block.
- * ----------------
+/*
+ * Error-checking support for DROP commands
  */
- /* we have to use IF because of the 'break' */
-#define CHECK_IF_ABORTED() \
-if (1) \
-{ \
-	if (IsAbortedTransactionBlockState()) \
-	{ \
-		elog(NOTICE, "current transaction is aborted, " \
-			 "queries ignored until end of transaction block"); \
-		commandTag = "*ABORT STATE*"; \
-		break; \
-	} \
-} else
+
+struct kindstrings
+{
+	char		kind;
+	char	   *indef_article;
+	char	   *name;
+	char	   *command;
+};
+
+static struct kindstrings kindstringarray[] = {
+	{RELKIND_RELATION, "a", "table", "TABLE"},
+	{RELKIND_SEQUENCE, "a", "sequence", "SEQUENCE"},
+	{RELKIND_VIEW, "a", "view", "VIEW"},
+	{RELKIND_INDEX, "an", "index", "INDEX"},
+	{'\0', "a", "???", "???"}
+};
+
+
+static void
+DropErrorMsg(char *relname, char wrongkind, char rightkind)
+{
+	struct kindstrings *rentry;
+	struct kindstrings *wentry;
+
+	for (rentry = kindstringarray; rentry->kind != '\0'; rentry++)
+		if (rentry->kind == rightkind)
+			break;
+	Assert(rentry->kind != '\0');
+
+	for (wentry = kindstringarray; wentry->kind != '\0'; wentry++)
+		if (wentry->kind == wrongkind)
+			break;
+	/* wrongkind could be something we don't have in our table... */
+	if (wentry->kind != '\0')
+		elog(ERROR, "\"%s\" is not %s %s. Use DROP %s to remove %s %s",
+			 relname, rentry->indef_article, rentry->name,
+			 wentry->command, wentry->indef_article, wentry->name);
+	else
+		elog(ERROR, "\"%s\" is not %s %s",
+			 relname, rentry->indef_article, rentry->name);
+}
+
+static void
+CheckDropPermissions(char *name, char rightkind)
+{
+	struct kindstrings *rentry;
+	HeapTuple	tuple;
+	Form_pg_class classform;
+
+	for (rentry = kindstringarray; rentry->kind != '\0'; rentry++)
+		if (rentry->kind == rightkind)
+			break;
+	Assert(rentry->kind != '\0');
+
+	tuple = SearchSysCache(RELNAME,
+						   PointerGetDatum(name),
+						   0, 0, 0);
+	if (!HeapTupleIsValid(tuple))
+		elog(ERROR, "%s \"%s\" does not exist", rentry->name, name);
+
+	classform = (Form_pg_class) GETSTRUCT(tuple);
+
+	if (classform->relkind != rightkind)
+		DropErrorMsg(name, classform->relkind, rightkind);
+
+	if (!pg_ownercheck(GetUserId(), name, RELNAME))
+		elog(ERROR, "you do not own %s \"%s\"",
+			 rentry->name, name);
+
+	if (!allowSystemTableMods && IsSystemRelationName(name))
+		elog(ERROR, "%s \"%s\" is a system %s",
+			 rentry->name, name, rentry->name);
+
+	ReleaseSysCache(tuple);
+}
+
 
 /* ----------------
  *		general utility function invoker
@@ -75,9 +139,6 @@ ProcessUtility(Node *parsetree,
 	char	   *commandTag = NULL;
 	char	   *relname;
 	char	   *relationName;
-	char	   *userName;
-
-	userName = GetPgUserName();
 
 	switch (nodeTag(parsetree))
 	{
@@ -93,18 +154,17 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->command)
 				{
 					case BEGIN_TRANS:
-						PS_SET_STATUS(commandTag = "BEGIN");
-						CHECK_IF_ABORTED();
+						set_ps_display(commandTag = "BEGIN");
 						BeginTransactionBlock();
 						break;
 
 					case COMMIT:
-						PS_SET_STATUS(commandTag = "COMMIT");
+						set_ps_display(commandTag = "COMMIT");
 						EndTransactionBlock();
 						break;
 
 					case ROLLBACK:
-						PS_SET_STATUS(commandTag = "ROLLBACK");
+						set_ps_display(commandTag = "ROLLBACK");
 						UserAbortTransactionBlock();
 						break;
 				}
@@ -119,8 +179,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ClosePortalStmt *stmt = (ClosePortalStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CLOSE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CLOSE");
 
 				PerformPortalClose(stmt->portalname, dest);
 			}
@@ -133,8 +192,7 @@ ProcessUtility(Node *parsetree,
 				bool		forward;
 				int			count;
 
-				PS_SET_STATUS(commandTag = (stmt->ismove) ? "MOVE" : "FETCH");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = (stmt->ismove) ? "MOVE" : "FETCH");
 
 				SetQuerySnapshot();
 
@@ -156,52 +214,79 @@ ProcessUtility(Node *parsetree,
 			 *
 			 */
 		case T_CreateStmt:
-			PS_SET_STATUS(commandTag = "CREATE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE");
 
 			DefineRelation((CreateStmt *) parsetree, RELKIND_RELATION);
+
+			/*
+			 * Let AlterTableCreateToastTable decide if this one needs a
+			 * secondary relation too.
+			 */
+			CommandCounterIncrement();
+			AlterTableCreateToastTable(((CreateStmt *) parsetree)->relname,
+									   true);
 			break;
 
 		case T_DropStmt:
 			{
 				DropStmt   *stmt = (DropStmt *) parsetree;
-				List	   *args = stmt->relNames;
+				List	   *args = stmt->names;
 				List	   *arg;
 
-				PS_SET_STATUS(commandTag = "DROP");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "DROP");
 
-				/* check as much as we can before we start dropping ... */
-				foreach(arg, args)
-				{
-					Relation	rel;
-
-					relname = strVal(lfirst(arg));
-					if (!allowSystemTableMods && IsSystemRelationName(relname))
-						elog(ERROR, "class \"%s\" is a system catalog",
-							 relname);
-					rel = heap_openr(relname, AccessExclusiveLock);
-					if (stmt->sequence &&
-						rel->rd_rel->relkind != RELKIND_SEQUENCE)
-						elog(ERROR, "Use DROP TABLE to drop table '%s'",
-							 relname);
-					if (!(stmt->sequence) &&
-						rel->rd_rel->relkind == RELKIND_SEQUENCE)
-						elog(ERROR, "Use DROP SEQUENCE to drop sequence '%s'",
-							 relname);
-					/* close rel, but keep lock until end of xact */
-					heap_close(rel, NoLock);
-#ifndef NO_SECURITY
-					if (!pg_ownercheck(userName, relname, RELNAME))
-						elog(ERROR, "you do not own class \"%s\"",
-							 relname);
-#endif
-				}
-				/* OK, terminate 'em all */
 				foreach(arg, args)
 				{
 					relname = strVal(lfirst(arg));
-					RemoveRelation(relname);
+
+					switch (stmt->removeType)
+					{
+						case DROP_TABLE:
+							CheckDropPermissions(relname, RELKIND_RELATION);
+							RemoveRelation(relname);
+							break;
+
+						case DROP_SEQUENCE:
+							CheckDropPermissions(relname, RELKIND_SEQUENCE);
+							RemoveRelation(relname);
+							break;
+
+						case DROP_VIEW:
+							CheckDropPermissions(relname, RELKIND_VIEW);
+							RemoveView(relname);
+							break;
+
+						case DROP_INDEX:
+							CheckDropPermissions(relname, RELKIND_INDEX);
+							RemoveIndex(relname);
+							break;
+
+						case DROP_RULE:
+							{
+								char	   *rulename = relname;
+								int			aclcheck_result;
+
+								relationName = RewriteGetRuleEventRel(rulename);
+								aclcheck_result = pg_aclcheck(relationName, GetUserId(), ACL_RU);
+								if (aclcheck_result != ACLCHECK_OK)
+									elog(ERROR, "%s: %s", relationName,
+										 aclcheck_error_strings[aclcheck_result]);
+								RemoveRewriteRule(rulename);
+							}
+							break;
+
+						case DROP_TYPE_P:
+							/* RemoveType does its own permissions checks */
+							RemoveType(relname);
+							break;
+					}
+
+					/*
+					 * Make sure subsequent loop iterations will see
+					 * results of this one; needed if removing multiple
+					 * rules for same table, for example.
+					 */
+					CommandCounterIncrement();
 				}
 			}
 			break;
@@ -210,8 +295,7 @@ ProcessUtility(Node *parsetree,
 			{
 				Relation	rel;
 
-				PS_SET_STATUS(commandTag = "TRUNCATE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "TRUNCATE");
 
 				relname = ((TruncateStmt *) parsetree)->relName;
 				if (!allowSystemTableMods && IsSystemRelationName(relname))
@@ -223,39 +307,36 @@ ProcessUtility(Node *parsetree,
 				if (rel->rd_rel->relkind == RELKIND_SEQUENCE)
 					elog(ERROR, "TRUNCATE cannot be used on sequences. '%s' is a sequence",
 						 relname);
+				if (rel->rd_rel->relkind == RELKIND_VIEW)
+					elog(ERROR, "TRUNCATE cannot be used on views. '%s' is a sequence",
+						 relname);
 				heap_close(rel, NoLock);
 
-#ifndef NO_SECURITY
-				if (!pg_ownercheck(userName, relname, RELNAME))
+				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
 					elog(ERROR, "you do not own class \"%s\"", relname);
-#endif
 				TruncateRelation(relname);
 			}
 			break;
 
 		case T_CommentStmt:
 			{
-
 				CommentStmt *statement;
 
 				statement = ((CommentStmt *) parsetree);
 
-				PS_SET_STATUS(commandTag = "COMMENT");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "COMMENT");
+
 				CommentObject(statement->objtype, statement->objname,
 							  statement->objproperty, statement->objlist,
 							  statement->comment);
 			}
 			break;
 
-
-
 		case T_CopyStmt:
 			{
 				CopyStmt   *stmt = (CopyStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "COPY");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "COPY");
 
 				if (stmt->direction != FROM)
 					SetQuerySnapshot();
@@ -283,17 +364,14 @@ ProcessUtility(Node *parsetree,
 			{
 				RenameStmt *stmt = (RenameStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "ALTER");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "ALTER");
 
 				relname = stmt->relname;
 				if (!allowSystemTableMods && IsSystemRelationName(relname))
 					elog(ERROR, "ALTER TABLE: relation \"%s\" is a system catalog",
 						 relname);
-#ifndef NO_SECURITY
-				if (!pg_ownercheck(userName, relname, RELNAME))
+				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
 					elog(ERROR, "permission denied");
-#endif
 
 				/* ----------------
 				 *	XXX using len == 3 to tell the difference
@@ -306,29 +384,26 @@ ProcessUtility(Node *parsetree,
 				 */
 				if (stmt->column == NULL)
 				{
-					/* ----------------
-					 *		rename relation
+
+					/*
+					 * rename relation
 					 *
-					 *		Note: we also rename the "type" tuple
-					 *		corresponding to the relation.
-					 * ----------------
+					 * Note: we also rename the "type" tuple corresponding to
+					 * the relation.
 					 */
 					renamerel(relname,	/* old name */
 							  stmt->newname);	/* new name */
-					TypeRename(relname, /* old name */
-							   stmt->newname);	/* new name */
 				}
 				else
 				{
-					/* ----------------
-					 *		rename attribute
-					 * ----------------
+
+					/*
+					 * rename attribute
 					 */
 					renameatt(relname,	/* relname */
 							  stmt->column,		/* old att name */
 							  stmt->newname,	/* new att name */
-							  userName,
-							  stmt->inh);		/* recursive? */
+							  interpretInhOption(stmt->inhOpt));		/* recursive? */
 				}
 			}
 			break;
@@ -339,8 +414,7 @@ ProcessUtility(Node *parsetree,
 			{
 				AlterTableStmt *stmt = (AlterTableStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "ALTER");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "ALTER");
 
 				/*
 				 * Some or all of these functions are recursive to cover
@@ -349,19 +423,40 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->subtype)
 				{
 					case 'A':	/* ADD COLUMN */
-						AlterTableAddColumn(stmt->relname, stmt->inh, (ColumnDef *) stmt->def);
+						AlterTableAddColumn(stmt->relname,
+										interpretInhOption(stmt->inhOpt),
+											(ColumnDef *) stmt->def);
 						break;
 					case 'T':	/* ALTER COLUMN */
-						AlterTableAlterColumn(stmt->relname, stmt->inh, stmt->name, stmt->def);
+						AlterTableAlterColumn(stmt->relname,
+										interpretInhOption(stmt->inhOpt),
+											  stmt->name,
+											  stmt->def);
 						break;
 					case 'D':	/* ALTER DROP */
-						AlterTableDropColumn(stmt->relname, stmt->inh, stmt->name, stmt->behavior);
+						AlterTableDropColumn(stmt->relname,
+										interpretInhOption(stmt->inhOpt),
+											 stmt->name,
+											 stmt->behavior);
 						break;
 					case 'C':	/* ADD CONSTRAINT */
-						AlterTableAddConstraint(stmt->relname, stmt->inh, stmt->def);
+						AlterTableAddConstraint(stmt->relname,
+										interpretInhOption(stmt->inhOpt),
+												stmt->def);
 						break;
 					case 'X':	/* DROP CONSTRAINT */
-						AlterTableDropConstraint(stmt->relname, stmt->inh, stmt->name, stmt->behavior);
+						AlterTableDropConstraint(stmt->relname,
+										interpretInhOption(stmt->inhOpt),
+												 stmt->name,
+												 stmt->behavior);
+						break;
+					case 'E':	/* CREATE TOAST TABLE */
+						AlterTableCreateToastTable(stmt->relname,
+												   false);
+						break;
+					case 'U':	/* ALTER OWNER */
+						AlterTableOwner(stmt->relname,
+										stmt->name);
 						break;
 					default:	/* oops */
 						elog(ERROR, "T_AlterTableStmt: unknown subtype");
@@ -374,34 +469,10 @@ ProcessUtility(Node *parsetree,
 		case T_ChangeACLStmt:
 			{
 				ChangeACLStmt *stmt = (ChangeACLStmt *) parsetree;
-				List	   *i;
-				AclItem    *aip;
-				unsigned	modechg;
 
-				PS_SET_STATUS(commandTag = "CHANGE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CHANGE");
 
-				aip = stmt->aclitem;
-
-				modechg = stmt->modechg;
-				foreach(i, stmt->relNames)
-				{
-					Relation	rel;
-
-					relname = strVal(lfirst(i));
-					rel = heap_openr(relname, AccessExclusiveLock);
-					if (rel && rel->rd_rel->relkind == RELKIND_INDEX)
-						elog(ERROR, "\"%s\" is an index relation",
-							 relname);
-					/* close rel, but keep lock until end of xact */
-					heap_close(rel, NoLock);
-#ifndef NO_SECURITY
-					if (!pg_ownercheck(userName, relname, RELNAME))
-						elog(ERROR, "you do not own class \"%s\"",
-							 relname);
-#endif
-					ChangeAcl(relname, aip, modechg);
-				}
+				ExecuteChangeACLStmt(stmt);
 			}
 			break;
 
@@ -414,8 +485,7 @@ ProcessUtility(Node *parsetree,
 			{
 				DefineStmt *stmt = (DefineStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CREATE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CREATE");
 
 				switch (stmt->defType)
 				{
@@ -438,15 +508,15 @@ ProcessUtility(Node *parsetree,
 			{
 				ViewStmt   *stmt = (ViewStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CREATE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CREATE");
+
 				DefineView(stmt->viewname, stmt->query);		/* retrieve parsetree */
 			}
 			break;
 
 		case T_ProcedureStmt:	/* CREATE FUNCTION */
-			PS_SET_STATUS(commandTag = "CREATE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE");
+
 			CreateFunction((ProcedureStmt *) parsetree, dest);	/* everything */
 			break;
 
@@ -454,8 +524,8 @@ ProcessUtility(Node *parsetree,
 			{
 				IndexStmt  *stmt = (IndexStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CREATE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CREATE");
+
 				DefineIndex(stmt->relname,		/* relation name */
 							stmt->idxname,		/* index name */
 							stmt->accessMethod, /* am name */
@@ -473,21 +543,18 @@ ProcessUtility(Node *parsetree,
 				RuleStmt   *stmt = (RuleStmt *) parsetree;
 				int			aclcheck_result;
 
-#ifndef NO_SECURITY
 				relname = stmt->object->relname;
-				aclcheck_result = pg_aclcheck(relname, userName, ACL_RU);
+				aclcheck_result = pg_aclcheck(relname, GetUserId(), ACL_RU);
 				if (aclcheck_result != ACLCHECK_OK)
 					elog(ERROR, "%s: %s", relname, aclcheck_error_strings[aclcheck_result]);
-#endif
-				PS_SET_STATUS(commandTag = "CREATE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CREATE");
+
 				DefineQueryRewrite(stmt);
 			}
 			break;
 
 		case T_CreateSeqStmt:
-			PS_SET_STATUS(commandTag = "CREATE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE");
 
 			DefineSequence((CreateSeqStmt *) parsetree);
 			break;
@@ -496,8 +563,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ExtendStmt *stmt = (ExtendStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "EXTEND");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "EXTEND");
 
 				ExtendIndex(stmt->idxname,		/* index name */
 							(Expr *) stmt->whereClause, /* where */
@@ -505,75 +571,17 @@ ProcessUtility(Node *parsetree,
 			}
 			break;
 
-		case T_RemoveStmt:
-			{
-				RemoveStmt *stmt = (RemoveStmt *) parsetree;
-
-				PS_SET_STATUS(commandTag = "DROP");
-				CHECK_IF_ABORTED();
-
-				switch (stmt->removeType)
-				{
-					case INDEX:
-						relname = stmt->name;
-						if (!allowSystemTableMods && IsSystemRelationName(relname))
-							elog(ERROR, "class \"%s\" is a system catalog index",
-								 relname);
-#ifndef NO_SECURITY
-						if (!pg_ownercheck(userName, relname, RELNAME))
-							elog(ERROR, "%s: %s", relname, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-#endif
-						RemoveIndex(relname);
-						break;
-					case RULE:
-						{
-							char	   *rulename = stmt->name;
-							int			aclcheck_result;
-
-#ifndef NO_SECURITY
-
-							relationName = RewriteGetRuleEventRel(rulename);
-							aclcheck_result = pg_aclcheck(relationName, userName, ACL_RU);
-							if (aclcheck_result != ACLCHECK_OK)
-								elog(ERROR, "%s: %s", relationName, aclcheck_error_strings[aclcheck_result]);
-#endif
-							RemoveRewriteRule(rulename);
-						}
-						break;
-					case TYPE_P:
-#ifndef NO_SECURITY
-						/* XXX moved to remove.c */
-#endif
-						RemoveType(stmt->name);
-						break;
-					case VIEW:
-						{
-							char	   *viewName = stmt->name;
-							char	   *ruleName;
-
-#ifndef NO_SECURITY
-
-							ruleName = MakeRetrieveViewRuleName(viewName);
-							relationName = RewriteGetRuleEventRel(ruleName);
-							if (!pg_ownercheck(userName, relationName, RELNAME))
-								elog(ERROR, "%s: %s", relationName, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-							pfree(ruleName);
-#endif
-							RemoveView(viewName);
-						}
-						break;
-				}
-				break;
-			}
-			break;
-
 		case T_RemoveAggrStmt:
 			{
 				RemoveAggrStmt *stmt = (RemoveAggrStmt *) parsetree;
+				char	   *typename = (char *) NULL;
 
-				PS_SET_STATUS(commandTag = "DROP");
-				CHECK_IF_ABORTED();
-				RemoveAggregate(stmt->aggname, stmt->aggtype);
+				set_ps_display(commandTag = "DROP");
+
+				if (stmt->aggtype != NULL)
+					typename = TypeNameToInternalName((TypeName *) stmt->aggtype);
+
+				RemoveAggregate(stmt->aggname, typename);
 			}
 			break;
 
@@ -581,28 +589,28 @@ ProcessUtility(Node *parsetree,
 			{
 				RemoveFuncStmt *stmt = (RemoveFuncStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "DROP");
-				CHECK_IF_ABORTED();
-				RemoveFunction(stmt->funcname,
-							   length(stmt->args),
-							   stmt->args);
+				set_ps_display(commandTag = "DROP");
+
+				RemoveFunction(stmt->funcname, stmt->args);
 			}
 			break;
 
 		case T_RemoveOperStmt:
 			{
 				RemoveOperStmt *stmt = (RemoveOperStmt *) parsetree;
-				char	   *type1 = (char *) NULL;
-				char	   *type2 = (char *) NULL;
+				TypeName   *typenode1 = (TypeName *) lfirst(stmt->args);
+				TypeName   *typenode2 = (TypeName *) lsecond(stmt->args);
+				char	   *typename1 = (char *) NULL;
+				char	   *typename2 = (char *) NULL;
 
-				PS_SET_STATUS(commandTag = "DROP");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "DROP");
 
-				if (lfirst(stmt->args) != NULL)
-					type1 = strVal(lfirst(stmt->args));
-				if (lsecond(stmt->args) != NULL)
-					type2 = strVal(lsecond(stmt->args));
-				RemoveOperator(stmt->opname, type1, type2);
+				if (typenode1 != NULL)
+					typename1 = TypeNameToInternalName(typenode1);
+				if (typenode2 != NULL)
+					typename2 = TypeNameToInternalName(typenode2);
+
+				RemoveOperator(stmt->opname, typename1, typename2);
 			}
 			break;
 
@@ -614,9 +622,10 @@ ProcessUtility(Node *parsetree,
 			{
 				CreatedbStmt *stmt = (CreatedbStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CREATE DATABASE");
-				CHECK_IF_ABORTED();
-				createdb(stmt->dbname, stmt->dbpath, stmt->encoding);
+				set_ps_display(commandTag = "CREATE DATABASE");
+
+				createdb(stmt->dbname, stmt->dbpath,
+						 stmt->dbtemplate, stmt->encoding);
 			}
 			break;
 
@@ -624,8 +633,8 @@ ProcessUtility(Node *parsetree,
 			{
 				DropdbStmt *stmt = (DropdbStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "DROP DATABASE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "DROP DATABASE");
+
 				dropdb(stmt->dbname);
 			}
 			break;
@@ -635,8 +644,7 @@ ProcessUtility(Node *parsetree,
 			{
 				NotifyStmt *stmt = (NotifyStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "NOTIFY");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "NOTIFY");
 
 				Async_Notify(stmt->relname);
 			}
@@ -646,8 +654,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ListenStmt *stmt = (ListenStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "LISTEN");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "LISTEN");
 
 				Async_Listen(stmt->relname, MyProcPid);
 			}
@@ -657,8 +664,7 @@ ProcessUtility(Node *parsetree,
 			{
 				UnlistenStmt *stmt = (UnlistenStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "UNLISTEN");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "UNLISTEN");
 
 				Async_Unlisten(stmt->relname, MyProcPid);
 			}
@@ -672,8 +678,7 @@ ProcessUtility(Node *parsetree,
 			{
 				LoadStmt   *stmt = (LoadStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "LOAD");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "LOAD");
 
 				closeAllVfds(); /* probably not necessary... */
 				load_file(stmt->filename);
@@ -684,16 +689,22 @@ ProcessUtility(Node *parsetree,
 			{
 				ClusterStmt *stmt = (ClusterStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "CLUSTER");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "CLUSTER");
 
-				cluster(stmt->relname, stmt->indexname);
+				relname = stmt->relname;
+				if (IsSystemRelationName(relname))
+					elog(ERROR, "CLUSTER: relation \"%s\" is a system catalog",
+						 relname);
+				if (!pg_ownercheck(GetUserId(), relname, RELNAME))
+					elog(ERROR, "permission denied");
+
+				cluster(relname, stmt->indexname);
 			}
 			break;
 
 		case T_VacuumStmt:
-			PS_SET_STATUS(commandTag = "VACUUM");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "VACUUM");
+
 			vacuum(((VacuumStmt *) parsetree)->vacrel,
 				   ((VacuumStmt *) parsetree)->verbose,
 				   ((VacuumStmt *) parsetree)->analyze,
@@ -704,8 +715,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ExplainStmt *stmt = (ExplainStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "EXPLAIN");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "EXPLAIN");
 
 				ExplainQuery(stmt->query, stmt->verbose, dest);
 			}
@@ -720,8 +730,8 @@ ProcessUtility(Node *parsetree,
 			{
 				RecipeStmt *stmt = (RecipeStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "EXECUTE RECIPE");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "EXECUTE RECIPE");
+
 				beginRecipe(stmt);
 			}
 			break;
@@ -735,7 +745,7 @@ ProcessUtility(Node *parsetree,
 				VariableSetStmt *n = (VariableSetStmt *) parsetree;
 
 				SetPGVariable(n->name, n->value);
-				PS_SET_STATUS(commandTag = "SET VARIABLE");
+				set_ps_display(commandTag = "SET VARIABLE");
 			}
 			break;
 
@@ -744,7 +754,7 @@ ProcessUtility(Node *parsetree,
 				VariableShowStmt *n = (VariableShowStmt *) parsetree;
 
 				GetPGVariable(n->name);
-				PS_SET_STATUS(commandTag = "SHOW VARIABLE");
+				set_ps_display(commandTag = "SHOW VARIABLE");
 			}
 			break;
 
@@ -753,7 +763,7 @@ ProcessUtility(Node *parsetree,
 				VariableResetStmt *n = (VariableResetStmt *) parsetree;
 
 				ResetPGVariable(n->name);
-				PS_SET_STATUS(commandTag = "RESET VARIABLE");
+				set_ps_display(commandTag = "RESET VARIABLE");
 			}
 			break;
 
@@ -761,15 +771,13 @@ ProcessUtility(Node *parsetree,
 			 * ******************************** TRIGGER statements *******************************
 			 */
 		case T_CreateTrigStmt:
-			PS_SET_STATUS(commandTag = "CREATE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE");
 
 			CreateTrigger((CreateTrigStmt *) parsetree);
 			break;
 
 		case T_DropTrigStmt:
-			PS_SET_STATUS(commandTag = "DROP");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "DROP");
 
 			DropTrigger((DropTrigStmt *) parsetree);
 			break;
@@ -778,15 +786,13 @@ ProcessUtility(Node *parsetree,
 			 * ************* PROCEDURAL LANGUAGE statements *****************
 			 */
 		case T_CreatePLangStmt:
-			PS_SET_STATUS(commandTag = "CREATE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE");
 
 			CreateProceduralLanguage((CreatePLangStmt *) parsetree);
 			break;
 
 		case T_DropPLangStmt:
-			PS_SET_STATUS(commandTag = "DROP");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "DROP");
 
 			DropProceduralLanguage((DropPLangStmt *) parsetree);
 			break;
@@ -796,67 +802,68 @@ ProcessUtility(Node *parsetree,
 			 *
 			 */
 		case T_CreateUserStmt:
-			PS_SET_STATUS(commandTag = "CREATE USER");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE USER");
 
 			CreateUser((CreateUserStmt *) parsetree);
 			break;
 
 		case T_AlterUserStmt:
-			PS_SET_STATUS(commandTag = "ALTER USER");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "ALTER USER");
 
 			AlterUser((AlterUserStmt *) parsetree);
 			break;
 
 		case T_DropUserStmt:
-			PS_SET_STATUS(commandTag = "DROP USER");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "DROP USER");
 
 			DropUser((DropUserStmt *) parsetree);
 			break;
 
 		case T_LockStmt:
-			PS_SET_STATUS(commandTag = "LOCK TABLE");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "LOCK TABLE");
 
 			LockTableCommand((LockStmt *) parsetree);
 			break;
 
 		case T_ConstraintsSetStmt:
-			PS_SET_STATUS(commandTag = "SET CONSTRAINTS");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "SET CONSTRAINTS");
 
 			DeferredTriggerSetState((ConstraintsSetStmt *) parsetree);
 			break;
 
 		case T_CreateGroupStmt:
-			PS_SET_STATUS(commandTag = "CREATE GROUP");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "CREATE GROUP");
 
 			CreateGroup((CreateGroupStmt *) parsetree);
 			break;
 
 		case T_AlterGroupStmt:
-			PS_SET_STATUS(commandTag = "ALTER GROUP");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "ALTER GROUP");
 
 			AlterGroup((AlterGroupStmt *) parsetree, "ALTER GROUP");
 			break;
 
 		case T_DropGroupStmt:
-			PS_SET_STATUS(commandTag = "DROP GROUP");
-			CHECK_IF_ABORTED();
+			set_ps_display(commandTag = "DROP GROUP");
 
 			DropGroup((DropGroupStmt *) parsetree);
 			break;
 
+		case T_CheckPointStmt:
+			{
+				set_ps_display(commandTag = "CHECKPOINT");
+
+				if (!superuser())
+					elog(ERROR, "permission denied");
+				CreateCheckPoint(false);
+			}
+			break;
+
 		case T_ReindexStmt:
 			{
 				ReindexStmt *stmt = (ReindexStmt *) parsetree;
 
-				PS_SET_STATUS(commandTag = "REINDEX");
-				CHECK_IF_ABORTED();
+				set_ps_display(commandTag = "REINDEX");
 
 				switch (stmt->reindexType)
 				{
@@ -866,33 +873,31 @@ ProcessUtility(Node *parsetree,
 						{
 							if (!allowSystemTableMods && IsSystemRelationName(relname))
 								elog(ERROR, "\"%s\" is a system index. call REINDEX under standalone postgres with -O -P options",
-								 relname);
+									 relname);
 							if (!IsIgnoringSystemIndexes())
 								elog(ERROR, "\"%s\" is a system index. call REINDEX under standalone postgres with -P -O options",
-								 relname);
+									 relname);
 						}
-#ifndef NO_SECURITY
-						if (!pg_ownercheck(userName, relname, RELNAME))
+						if (!pg_ownercheck(GetUserId(), relname, RELNAME))
 							elog(ERROR, "%s: %s", relname, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-#endif
 						ReindexIndex(relname, stmt->force);
 						break;
 					case TABLE:
 						relname = (char *) stmt->name;
 						if (IsSystemRelationName(relname))
 						{
+#ifdef	OLD_FILE_NAMING
 							if (!allowSystemTableMods && IsSystemRelationName(relname))
 								elog(ERROR, "\"%s\" is a system table. call REINDEX under standalone postgres with -O -P options",
-								 relname);
+									 relname);
 							if (!IsIgnoringSystemIndexes())
 								elog(ERROR, "\"%s\" is a system table. call REINDEX under standalone postgres with -P -O options",
 
-								 relname);
+									 relname);
+#endif	 /* OLD_FILE_NAMING */
 						}
-#ifndef NO_SECURITY
-						if (!pg_ownercheck(userName, relname, RELNAME))
+						if (!pg_ownercheck(GetUserId(), relname, RELNAME))
 							elog(ERROR, "%s: %s", relname, aclcheck_error_strings[ACLCHECK_NOT_OWNER]);
-#endif
 						ReindexTable(relname, stmt->force);
 						break;
 					case DATABASE:
@@ -918,9 +923,8 @@ ProcessUtility(Node *parsetree,
 			break;
 	}
 
-	/* ----------------
-	 *	tell fe/be or whatever that we're done.
-	 * ----------------
+	/*
+	 * tell fe/be or whatever that we're done.
 	 */
 	EndCommand(commandTag, dest);
 }
