@@ -3,20 +3,21 @@
  * varlena.c
  *	  Functions for the variable-length built-in types.
  *
- * Portions Copyright (c) 1996-2000, PostgreSQL, Inc
+ * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/utils/adt/varlena.c,v 1.58 2000/04/12 17:15:52 momjian Exp $
+ *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/utils/adt/varlena.c,v 1.70 2001/05/03 19:00:36 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
-#include <ctype.h>
-
 #include "postgres.h"
 
+#include <ctype.h>
+
 #include "mb/pg_wchar.h"
+#include "miscadmin.h"
 #include "utils/builtins.h"
 
 static int	text_cmp(text *arg1, text *arg2);
@@ -41,33 +42,36 @@ static int	text_cmp(text *arg1, text *arg2);
  *				The input is scaned twice.
  *				The error checking of input is minimal.
  */
-bytea *
-byteain(char *inputText)
+Datum
+byteain(PG_FUNCTION_ARGS)
 {
+	char	   *inputText = PG_GETARG_CSTRING(0);
 	char	   *tp;
 	char	   *rp;
 	int			byte;
 	bytea	   *result;
 
-	if (inputText == NULL)
-		elog(ERROR, "Bad input string for type bytea");
-
 	for (byte = 0, tp = inputText; *tp != '\0'; byte++)
+	{
 		if (*tp++ == '\\')
 		{
 			if (*tp == '\\')
 				tp++;
-			else if (!isdigit(*tp++) ||
-					 !isdigit(*tp++) ||
-					 !isdigit(*tp++))
+			else if (!isdigit((unsigned char) *tp++) ||
+					 !isdigit((unsigned char) *tp++) ||
+					 !isdigit((unsigned char) *tp++))
 				elog(ERROR, "Bad input string for type bytea");
 		}
-	tp = inputText;
+	}
+
 	byte += VARHDRSZ;
 	result = (bytea *) palloc(byte);
-	result->vl_len = byte;		/* varlena? */
+	result->vl_len = byte;		/* set varlena length */
+
+	tp = inputText;
 	rp = result->vl_dat;
 	while (*tp != '\0')
+	{
 		if (*tp != '\\' || *++tp == '\\')
 			*rp++ = *tp++;
 		else
@@ -78,7 +82,9 @@ byteain(char *inputText)
 			byte <<= 3;
 			*rp++ = byte + VAL(*tp++);
 		}
-	return result;
+	}
+
+	PG_RETURN_BYTEA_P(result);
 }
 
 /*
@@ -89,9 +95,10 @@ byteain(char *inputText)
  *
  *		NULL vlena should be an error--returning string with NULL for now.
  */
-char *
-byteaout(bytea *vlena)
+Datum
+byteaout(PG_FUNCTION_ARGS)
 {
+	bytea	   *vlena = PG_GETARG_BYTEA_P(0);
 	char	   *result;
 	char	   *vp;
 	char	   *rp;
@@ -99,101 +106,88 @@ byteaout(bytea *vlena)
 	int			i;
 	int			len;
 
-	if (vlena == NULL)
-	{
-		result = (char *) palloc(2);
-		result[0] = '-';
-		result[1] = '\0';
-		return result;
-	}
-	vp = vlena->vl_dat;
 	len = 1;					/* empty string has 1 char */
+	vp = vlena->vl_dat;
 	for (i = vlena->vl_len - VARHDRSZ; i != 0; i--, vp++)
+	{
 		if (*vp == '\\')
 			len += 2;
-		else if (isascii(*vp) && isprint(*vp))
+		else if (isprint((unsigned char) *vp))
 			len++;
 		else
-			len += VARHDRSZ;
+			len += 4;
+	}
 	rp = result = (char *) palloc(len);
 	vp = vlena->vl_dat;
-	for (i = vlena->vl_len - VARHDRSZ; i != 0; i--)
+	for (i = vlena->vl_len - VARHDRSZ; i != 0; i--, vp++)
+	{
 		if (*vp == '\\')
 		{
-			vp++;
 			*rp++ = '\\';
 			*rp++ = '\\';
 		}
-		else if (isascii(*vp) && isprint(*vp))
-			*rp++ = *vp++;
+		else if (isprint((unsigned char) *vp))
+			*rp++ = *vp;
 		else
 		{
-			val = *vp++;
-			*rp = '\\';
-			rp += 3;
-			*rp-- = DIG(val & 07);
+			val = *vp;
+			rp[0] = '\\';
+			rp[3] = DIG(val & 07);
 			val >>= 3;
-			*rp-- = DIG(val & 07);
+			rp[2] = DIG(val & 07);
 			val >>= 3;
-			*rp = DIG(val & 03);
-			rp += 3;
+			rp[1] = DIG(val & 03);
+			rp += 4;
 		}
+	}
 	*rp = '\0';
-	return result;
+	PG_RETURN_CSTRING(result);
 }
 
 
 /*
  *		textin			- converts "..." to internal representation
  */
-text *
-textin(char *inputText)
+Datum
+textin(PG_FUNCTION_ARGS)
 {
+	char	   *inputText = PG_GETARG_CSTRING(0);
 	text	   *result;
 	int			len;
 
-	if (inputText == NULL)
-		return NULL;
-
 	len = strlen(inputText) + VARHDRSZ;
 	result = (text *) palloc(len);
-	VARSIZE(result) = len;
+	VARATT_SIZEP(result) = len;
 
-	memmove(VARDATA(result), inputText, len - VARHDRSZ);
+	memcpy(VARDATA(result), inputText, len - VARHDRSZ);
 
 #ifdef CYR_RECODE
 	convertstr(VARDATA(result), len - VARHDRSZ, 0);
 #endif
 
-	return result;
+	PG_RETURN_TEXT_P(result);
 }
 
 /*
  *		textout			- converts internal representation to "..."
  */
-char *
-textout(text *vlena)
+Datum
+textout(PG_FUNCTION_ARGS)
 {
+	text	   *t = PG_GETARG_TEXT_P(0);
 	int			len;
 	char	   *result;
 
-	if (vlena == NULL)
-	{
-		result = (char *) palloc(2);
-		result[0] = '-';
-		result[1] = '\0';
-		return result;
-	}
-	len = VARSIZE(vlena) - VARHDRSZ;
+	len = VARSIZE(t) - VARHDRSZ;
 	result = (char *) palloc(len + 1);
-	memmove(result, VARDATA(vlena), len);
+	memcpy(result, VARDATA(t), len);
 	result[len] = '\0';
 
 #ifdef CYR_RECODE
 	convertstr(result, len, 1);
 #endif
 
-	return result;
+	PG_RETURN_CSTRING(result);
 }
 
 
@@ -204,21 +198,17 @@ textout(text *vlena)
  *	  returns the logical length of a text*
  *	   (which is less than the VARSIZE of the text*)
  */
-int32
-textlen(text *t)
+Datum
+textlen(PG_FUNCTION_ARGS)
 {
+	text	   *t = PG_GETARG_TEXT_P(0);
+
 #ifdef MULTIBYTE
 	unsigned char *s;
 	int			len,
 				l,
 				wl;
 
-#endif
-
-	if (!PointerIsValid(t))
-		return 0;
-
-#ifdef MULTIBYTE
 	len = 0;
 	s = VARDATA(t);
 	l = VARSIZE(t) - VARHDRSZ;
@@ -229,30 +219,35 @@ textlen(text *t)
 		s += wl;
 		len++;
 	}
-	return (len);
+	PG_RETURN_INT32(len);
 #else
-	return VARSIZE(t) - VARHDRSZ;
+	PG_RETURN_INT32(VARSIZE(t) - VARHDRSZ);
 #endif
-
-}	/* textlen() */
+}
 
 /*
  * textoctetlen -
  *	  returns the physical length of a text*
  *	   (which is less than the VARSIZE of the text*)
+ *
+ * XXX is it actually appropriate to return the compressed length
+ * when the value is compressed?  It's not at all clear to me that
+ * this is what SQL92 has in mind ...
  */
-int32
-textoctetlen(text *t)
+Datum
+textoctetlen(PG_FUNCTION_ARGS)
 {
-	if (!PointerIsValid(t))
-		return 0;
+	struct varattrib *t = (struct varattrib *) PG_GETARG_RAW_VARLENA_P(0);
 
-	return VARSIZE(t) - VARHDRSZ;
-}	/* textoctetlen() */
+	if (!VARATT_IS_EXTERNAL(t))
+		PG_RETURN_INT32(VARATT_SIZE(t) - VARHDRSZ);
+
+	PG_RETURN_INT32(t->va_content.va_external.va_extsize);
+}
 
 /*
  * textcat -
- *	  takes two text* and returns a text* that is the concatentation of
+ *	  takes two text* and returns a text* that is the concatenation of
  *	  the two.
  *
  * Rewritten by Sapa, sapa@hq.icb.chel.su. 8-Jul-96.
@@ -260,35 +255,30 @@ textoctetlen(text *t)
  * Allocate space for output in all cases.
  * XXX - thomas 1997-07-10
  */
-text *
-textcat(text *t1, text *t2)
+Datum
+textcat(PG_FUNCTION_ARGS)
 {
+	text	   *t1 = PG_GETARG_TEXT_P(0);
+	text	   *t2 = PG_GETARG_TEXT_P(1);
 	int			len1,
 				len2,
 				len;
-	char	   *ptr;
 	text	   *result;
-
-	if (!PointerIsValid(t1) || !PointerIsValid(t2))
-		return NULL;
+	char	   *ptr;
 
 	len1 = (VARSIZE(t1) - VARHDRSZ);
 	if (len1 < 0)
 		len1 = 0;
-	while (len1 > 0 && VARDATA(t1)[len1 - 1] == '\0')
-		len1--;
 
 	len2 = (VARSIZE(t2) - VARHDRSZ);
 	if (len2 < 0)
 		len2 = 0;
-	while (len2 > 0 && VARDATA(t2)[len2 - 1] == '\0')
-		len2--;
 
 	len = len1 + len2 + VARHDRSZ;
-	result = palloc(len);
+	result = (text *) palloc(len);
 
 	/* Set size of result string... */
-	VARSIZE(result) = len;
+	VARATT_SIZEP(result) = len;
 
 	/* Fill data field of result string... */
 	ptr = VARDATA(result);
@@ -297,8 +287,8 @@ textcat(text *t1, text *t2)
 	if (len2 > 0)
 		memcpy(ptr + len1, VARDATA(t2), len2);
 
-	return result;
-}	/* textcat() */
+	PG_RETURN_TEXT_P(result);
+}
 
 /*
  * text_substr()
@@ -323,9 +313,12 @@ textcat(text *t1, text *t2)
  * Formerly returned the entire string; now returns a portion.
  * - Thomas Lockhart 1998-12-10
  */
-text *
-text_substr(text *string, int32 m, int32 n)
+Datum
+text_substr(PG_FUNCTION_ARGS)
 {
+	text	   *string = PG_GETARG_TEXT_P(0);
+	int32		m = PG_GETARG_INT32(1);
+	int32		n = PG_GETARG_INT32(2);
 	text	   *ret;
 	int			len;
 
@@ -335,9 +328,6 @@ text_substr(text *string, int32 m, int32 n)
 
 #endif
 
-	if (string == (text *) NULL)
-		return string;
-
 	len = VARSIZE(string) - VARHDRSZ;
 #ifdef MULTIBYTE
 	len = pg_mbstrlen_with_len(VARDATA(string), len);
@@ -374,13 +364,14 @@ text_substr(text *string, int32 m, int32 n)
 		p += pg_mblen(p);
 	n = p - (VARDATA(string) + m);
 #endif
+
 	ret = (text *) palloc(VARHDRSZ + n);
-	VARSIZE(ret) = VARHDRSZ + n;
+	VARATT_SIZEP(ret) = VARHDRSZ + n;
 
 	memcpy(VARDATA(ret), VARDATA(string) + m, n);
 
-	return ret;
-}	/* text_substr() */
+	PG_RETURN_TEXT_P(ret);
+}
 
 /*
  * textpos -
@@ -392,9 +383,11 @@ text_substr(text *string, int32 m, int32 n)
  * Added multi-byte support.
  * - Tatsuo Ishii 1998-4-21
  */
-int32
-textpos(text *t1, text *t2)
+Datum
+textpos(PG_FUNCTION_ARGS)
 {
+	text	   *t1 = PG_GETARG_TEXT_P(0);
+	text	   *t2 = PG_GETARG_TEXT_P(1);
 	int			pos;
 	int			px,
 				p;
@@ -409,11 +402,8 @@ textpos(text *t1, text *t2)
 
 #endif
 
-	if (!PointerIsValid(t1) || !PointerIsValid(t2))
-		return 0;
-
-	if (VARSIZE(t2) <= 0)
-		return 1;
+	if (VARSIZE(t2) <= VARHDRSZ)
+		PG_RETURN_INT32(1);		/* result for empty pattern */
 
 	len1 = (VARSIZE(t1) - VARHDRSZ);
 	len2 = (VARSIZE(t2) - VARHDRSZ);
@@ -447,43 +437,7 @@ textpos(text *t1, text *t2)
 	pfree(ps1);
 	pfree(ps2);
 #endif
-	return pos;
-}	/* textpos() */
-
-/*
- *		texteq			- returns 1 iff arguments are equal
- *		textne			- returns 1 iff arguments are not equal
- */
-bool
-texteq(text *arg1, text *arg2)
-{
-	int			len;
-	char	   *a1p,
-			   *a2p;
-
-	if (arg1 == NULL || arg2 == NULL)
-		return (bool) NULL;
-	if ((len = arg1->vl_len) != arg2->vl_len)
-		return (bool) 0;
-	a1p = arg1->vl_dat;
-	a2p = arg2->vl_dat;
-
-	/*
-	 * Varlenas are stored as the total size (data + size variable)
-	 * followed by the data. Use VARHDRSZ instead of explicit sizeof() -
-	 * thomas 1997-07-10
-	 */
-	len -= VARHDRSZ;
-	while (len-- != 0)
-		if (*a1p++ != *a2p++)
-			return (bool) 0;
-	return (bool) 1;
-}	/* texteq() */
-
-bool
-textne(text *arg1, text *arg2)
-{
-	return (bool) !texteq(arg1, arg2);
+	PG_RETURN_INT32(pos);
 }
 
 /* varstr_cmp()
@@ -500,8 +454,8 @@ varstr_cmp(char *arg1, int len1, char *arg2, int len2)
 			   *a2p;
 
 #ifdef USE_LOCALE
-	a1p = (unsigned char *) palloc(len1 + 1);
-	a2p = (unsigned char *) palloc(len2 + 1);
+	a1p = (char *) palloc(len1 + 1);
+	a2p = (char *) palloc(len2 + 1);
 
 	memcpy(a1p, arg1, len1);
 	*(a1p + len1) = '\0';
@@ -524,15 +478,11 @@ varstr_cmp(char *arg1, int len1, char *arg2, int len2)
 #endif
 
 	return result;
-}	/* varstr_cmp() */
+}
 
 
 /* text_cmp()
- * Comparison function for text strings.
- * Includes locale support, but must copy strings to temporary memory
- *	to allow null-termination for inputs to strcoll().
- * XXX HACK code for textlen() indicates that there can be embedded nulls
- *	but it appears that most routines (incl. this one) assume not! - tgl 97/04/07
+ * Internal comparison function for text strings.
  * Returns -1, 0 or 1
  */
 static int
@@ -543,9 +493,6 @@ text_cmp(text *arg1, text *arg2)
 	int			len1,
 				len2;
 
-	if (arg1 == NULL || arg2 == NULL)
-		return (bool) FALSE;
-
 	a1p = VARDATA(arg1);
 	a2p = VARDATA(arg2);
 
@@ -553,68 +500,152 @@ text_cmp(text *arg1, text *arg2)
 	len2 = VARSIZE(arg2) - VARHDRSZ;
 
 	return varstr_cmp(a1p, len1, a2p, len2);
-}	/* text_cmp() */
+}
 
-/* text_lt()
- * Comparison function for text strings.
+/*
+ * Comparison functions for text strings.
+ *
+ * Note: btree indexes need these routines not to leak memory; therefore,
+ * be careful to free working copies of toasted datums.  Most places don't
+ * need to be so careful.
  */
-bool
-text_lt(text *arg1, text *arg2)
-{
-	return (bool) (text_cmp(arg1, arg2) < 0);
-}	/* text_lt() */
 
-/* text_le()
- * Comparison function for text strings.
- */
-bool
-text_le(text *arg1, text *arg2)
+Datum
+texteq(PG_FUNCTION_ARGS)
 {
-	return (bool) (text_cmp(arg1, arg2) <= 0);
-}	/* text_le() */
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
 
-bool
-text_gt(text *arg1, text *arg2)
-{
-	return (bool) !text_le(arg1, arg2);
+	/* fast path for different-length inputs */
+	if (VARSIZE(arg1) != VARSIZE(arg2))
+		result = false;
+	else
+		result = (text_cmp(arg1, arg2) == 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
 }
 
-bool
-text_ge(text *arg1, text *arg2)
+Datum
+textne(PG_FUNCTION_ARGS)
 {
-	return (bool) !text_lt(arg1, arg2);
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
+
+	/* fast path for different-length inputs */
+	if (VARSIZE(arg1) != VARSIZE(arg2))
+		result = true;
+	else
+		result = (text_cmp(arg1, arg2) != 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
 }
 
-text *
-text_larger(text *arg1, text *arg2)
+Datum
+text_lt(PG_FUNCTION_ARGS)
 {
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
+
+	result = (text_cmp(arg1, arg2) < 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+text_le(PG_FUNCTION_ARGS)
+{
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
+
+	result = (text_cmp(arg1, arg2) <= 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+text_gt(PG_FUNCTION_ARGS)
+{
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
+
+	result = (text_cmp(arg1, arg2) > 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+text_ge(PG_FUNCTION_ARGS)
+{
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	bool		result;
+
+	result = (text_cmp(arg1, arg2) >= 0);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+bttextcmp(PG_FUNCTION_ARGS)
+{
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
+	int32		result;
+
+	result = text_cmp(arg1, arg2);
+
+	PG_FREE_IF_COPY(arg1, 0);
+	PG_FREE_IF_COPY(arg2, 1);
+
+	PG_RETURN_INT32(result);
+}
+
+
+Datum
+text_larger(PG_FUNCTION_ARGS)
+{
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
 	text	   *result;
-	text	   *temp;
 
-	temp = ((text_cmp(arg1, arg2) <= 0) ? arg2 : arg1);
+	result = ((text_cmp(arg1, arg2) > 0) ? arg1 : arg2);
 
-	/* Make a copy */
-
-	result = (text *) palloc(VARSIZE(temp));
-	memmove((char *) result, (char *) temp, VARSIZE(temp));
-
-	return (result);
+	PG_RETURN_TEXT_P(result);
 }
 
-text *
-text_smaller(text *arg1, text *arg2)
+Datum
+text_smaller(PG_FUNCTION_ARGS)
 {
+	text	   *arg1 = PG_GETARG_TEXT_P(0);
+	text	   *arg2 = PG_GETARG_TEXT_P(1);
 	text	   *result;
-	text	   *temp;
 
-	temp = ((text_cmp(arg1, arg2) > 0) ? arg2 : arg1);
+	result = ((text_cmp(arg1, arg2) < 0) ? arg1 : arg2);
 
-	/* Make a copy */
-
-	result = (text *) palloc(VARSIZE(temp));
-	memmove((char *) result, (char *) temp, VARSIZE(temp));
-
-	return (result);
+	PG_RETURN_TEXT_P(result);
 }
 
 /*-------------------------------------------------------------
@@ -623,32 +654,29 @@ text_smaller(text *arg1, text *arg2)
  * get the number of bytes contained in an instance of type 'bytea'
  *-------------------------------------------------------------
  */
-int32
-byteaoctetlen(bytea *v)
+Datum
+byteaoctetlen(PG_FUNCTION_ARGS)
 {
-	if (!PointerIsValid(v))
-		return 0;
+	bytea	   *v = PG_GETARG_BYTEA_P(0);
 
-	return VARSIZE(v) - VARHDRSZ;
+	PG_RETURN_INT32(VARSIZE(v) - VARHDRSZ);
 }
 
 /*-------------------------------------------------------------
  * byteaGetByte
  *
  * this routine treats "bytea" as an array of bytes.
- * It returns the Nth byte (a number between 0 and 255) or
- * it dies if the length of this array is less than n.
+ * It returns the Nth byte (a number between 0 and 255).
  *-------------------------------------------------------------
  */
-int32
-byteaGetByte(bytea *v, int32 n)
+Datum
+byteaGetByte(PG_FUNCTION_ARGS)
 {
+	bytea	   *v = PG_GETARG_BYTEA_P(0);
+	int32		n = PG_GETARG_INT32(1);
 	int			len;
 	int			byte;
 
-	if (!PointerIsValid(v))
-		return 0;
-
 	len = VARSIZE(v) - VARHDRSZ;
 
 	if (n < 0 || n >= len)
@@ -657,7 +685,7 @@ byteaGetByte(bytea *v, int32 n)
 
 	byte = ((unsigned char *) VARDATA(v))[n];
 
-	return (int32) byte;
+	PG_RETURN_INT32(byte);
 }
 
 /*-------------------------------------------------------------
@@ -665,21 +693,19 @@ byteaGetByte(bytea *v, int32 n)
  *
  * This routine treats a "bytea" type like an array of bits.
  * It returns the value of the Nth bit (0 or 1).
- * If 'n' is out of range, it dies!
  *
  *-------------------------------------------------------------
  */
-int32
-byteaGetBit(bytea *v, int32 n)
+Datum
+byteaGetBit(PG_FUNCTION_ARGS)
 {
+	bytea	   *v = PG_GETARG_BYTEA_P(0);
+	int32		n = PG_GETARG_INT32(1);
 	int			byteNo,
 				bitNo;
 	int			len;
 	int			byte;
 
-	if (!PointerIsValid(v))
-		return 0;
-
 	len = VARSIZE(v) - VARHDRSZ;
 
 	if (n < 0 || n >= len * 8)
@@ -692,9 +718,9 @@ byteaGetBit(bytea *v, int32 n)
 	byte = ((unsigned char *) VARDATA(v))[byteNo];
 
 	if (byte & (1 << bitNo))
-		return (int32) 1;
+		PG_RETURN_INT32(1);
 	else
-		return (int32) 0;
+		PG_RETURN_INT32(0);
 }
 
 /*-------------------------------------------------------------
@@ -705,15 +731,15 @@ byteaGetBit(bytea *v, int32 n)
  *
  *-------------------------------------------------------------
  */
-bytea *
-byteaSetByte(bytea *v, int32 n, int32 newByte)
+Datum
+byteaSetByte(PG_FUNCTION_ARGS)
 {
+	bytea	   *v = PG_GETARG_BYTEA_P(0);
+	int32		n = PG_GETARG_INT32(1);
+	int32		newByte = PG_GETARG_INT32(2);
 	int			len;
 	bytea	   *res;
 
-	if (!PointerIsValid(v))
-		return 0;
-
 	len = VARSIZE(v) - VARHDRSZ;
 
 	if (n < 0 || n >= len)
@@ -731,7 +757,7 @@ byteaSetByte(bytea *v, int32 n, int32 newByte)
 	 */
 	((unsigned char *) VARDATA(res))[n] = newByte;
 
-	return res;
+	PG_RETURN_BYTEA_P(res);
 }
 
 /*-------------------------------------------------------------
@@ -742,9 +768,12 @@ byteaSetByte(bytea *v, int32 n, int32 newByte)
  *
  *-------------------------------------------------------------
  */
-bytea *
-byteaSetBit(bytea *v, int32 n, int32 newBit)
+Datum
+byteaSetBit(PG_FUNCTION_ARGS)
 {
+	bytea	   *v = PG_GETARG_BYTEA_P(0);
+	int32		n = PG_GETARG_INT32(1);
+	int32		newBit = PG_GETARG_INT32(2);
 	bytea	   *res;
 	int			len;
 	int			oldByte,
@@ -752,9 +781,6 @@ byteaSetBit(bytea *v, int32 n, int32 newBit)
 	int			byteNo,
 				bitNo;
 
-	if (!PointerIsValid(v))
-		return NULL;
-
 	len = VARSIZE(v) - VARHDRSZ;
 
 	if (n < 0 || n >= len * 8)
@@ -771,50 +797,50 @@ byteaSetBit(bytea *v, int32 n, int32 newBit)
 		elog(ERROR, "byteaSetBit: new bit must be 0 or 1");
 
 	/*
-	 * get the byte where the bit we want is stored.
+	 * Make a copy of the original varlena.
 	 */
-	oldByte = byteaGetByte(v, byteNo);
+	res = (bytea *) palloc(VARSIZE(v));
+	memcpy((char *) res, (char *) v, VARSIZE(v));
 
 	/*
-	 * calculate the new value for that byte
+	 * Update the byte.
 	 */
+	oldByte = ((unsigned char *) VARDATA(res))[byteNo];
+
 	if (newBit == 0)
 		newByte = oldByte & (~(1 << bitNo));
 	else
 		newByte = oldByte | (1 << bitNo);
 
-	/*
-	 * NOTE: 'byteaSetByte' creates a copy of 'v' & sets the byte.
-	 */
-	res = byteaSetByte(v, byteNo, newByte);
+	((unsigned char *) VARDATA(res))[byteNo] = newByte;
 
-	return res;
+	PG_RETURN_BYTEA_P(res);
 }
 
 
 /* text_name()
- * Converts a text() type to a NameData type.
+ * Converts a text type to a Name type.
  */
-NameData   *
-text_name(text *s)
+Datum
+text_name(PG_FUNCTION_ARGS)
 {
-	NameData   *result;
+	text	   *s = PG_GETARG_TEXT_P(0);
+	Name		result;
 	int			len;
 
-	if (s == NULL)
-		return NULL;
-
 	len = VARSIZE(s) - VARHDRSZ;
-	if (len > NAMEDATALEN)
-		len = NAMEDATALEN;
+
+	/* Truncate oversize input */
+	if (len >= NAMEDATALEN)
+		len = NAMEDATALEN - 1;
 
 #ifdef STRINGDEBUG
 	printf("text- convert string length %d (%d) ->%d\n",
 		   VARSIZE(s) - VARHDRSZ, VARSIZE(s), len);
 #endif
 
-	result = palloc(NAMEDATALEN);
-	StrNCpy(NameStr(*result), VARDATA(s), NAMEDATALEN);
+	result = (Name) palloc(NAMEDATALEN);
+	memcpy(NameStr(*result), VARDATA(s), len);
 
 	/* now null pad to full length... */
 	while (len < NAMEDATALEN)
@@ -823,21 +849,19 @@ text_name(text *s)
 		len++;
 	}
 
-	return result;
-}	/* text_name() */
+	PG_RETURN_NAME(result);
+}
 
 /* name_text()
- * Converts a NameData type to a text type.
+ * Converts a Name type to a text type.
  */
-text *
-name_text(NameData *s)
+Datum
+name_text(PG_FUNCTION_ARGS)
 {
+	Name		s = PG_GETARG_NAME(0);
 	text	   *result;
 	int			len;
 
-	if (s == NULL)
-		return NULL;
-
 	len = strlen(NameStr(*s));
 
 #ifdef STRINGDEBUG
@@ -846,8 +870,8 @@ name_text(NameData *s)
 #endif
 
 	result = palloc(VARHDRSZ + len);
-	strncpy(VARDATA(result), NameStr(*s), len);
-	VARSIZE(result) = len + VARHDRSZ;
+	VARATT_SIZEP(result) = VARHDRSZ + len;
+	memcpy(VARDATA(result), NameStr(*s), len);
 
-	return result;
-}	/* name_text() */
+	PG_RETURN_TEXT_P(result);
+}
